/************************************************************************/
/*                                                                      */
/*      Tile Tabs WE - Generic WebExtension - Background Page           */
/*                                                                      */
/*      Javascript for Background Page                                  */
/*                                                                      */
/*      Last Edit - 03 Mar 2018                                         */
/*                                                                      */
/*      Copyright (C) 2015-2018 DW-dev                                  */
/*                                                                      */
/*      Distributed under the GNU General Public License version 2      */
/*      See LICENCE.txt file and http://www.gnu.org/licenses/           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Refer to Google Chrome developer documentation:                     */
/*                                                                      */
/*  https://developer.chrome.com/extensions/overview                    */
/*  https://developer.chrome.com/extensions/content_scripts             */
/*  https://developer.chrome.com/extensions/messaging                   */
/*  https://developer.chrome.com/extensions/optionsV2                   */
/*  https://developer.chrome.com/extensions/contentSecurityPolicy       */
/*                                                                      */
/*  https://developer.chrome.com/extensions/manifest                    */
/*  https://developer.chrome.com/extensions/declare_permissions         */
/*  https://developer.chrome.com/extensions/match_patterns              */
/*                                                                      */
/*  https://developer.chrome.com/extensions/browserAction               */
/*  https://developer.chrome.com/extensions/commands                    */
/*  https://developer.chrome.com/extensions/contextMenus                */
/*  https://developer.chrome.com/extensions/extension                   */
/*  https://developer.chrome.com/extensions/notifications               */
/*  https://developer.chrome.com/extensions/runtime                     */
/*  https://developer.chrome.com/extensions/storage                     */
/*  https://developer.chrome.com/extensions/tabs                        */
/*  https://developer.chrome.com/extensions/windows                     */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Chrome API Limitations                                              */
/*                                                                      */
/*  1. chrome.windows.create creates black window if tabId and url      */
/*     parameters are both null for popup window.                       */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Firefox API Limitations                                             */
/*                                                                      */
/*  1. chrome.windows.create does not support focused parameter for     */
/*     normal or popup window. (not a problem for current design)       */
/*                                                                      */
/*  2. chrome.windows.create does not support left/top parameters       */
/*     for popup window, which is always centred on the screen.         */
/*     (see Bug 1275222) (workaround is to update window position)      */
/*                                                                      */
/*  3. chrome.windows.create fails with 'browser.ownerGlobal is null'   */
/*     exception, if the parent window is in maximized state.           */
/*     (workaround is to convert window to normal state)                */
/*                                                                      */
/*  4. On Windows 10, if windows are without Title Bar, then            */
/*     toggling popup window to normal window causes black areas        */
/*     (see Bug 1273153)  (workaround is to update window position)     */
/*                                                                      */
/*  5. On Windows 10, if windows are without Title Bar, then:           */
/*     - normal windows are slightly larger than with Title Bar         */
/*     - popup windows are slightly smaller than normal windows         */
/*     (see Bug 1284273) (user can set other margins options)           */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/*  Notes on Tile Tabs WE Operation                                     */
/*                                                                      */
/*  1. The basic idea is to provide 'split browser' functionality,      */
/*     to allow viewing of multiple tabs in a tiled layout within       */
/*     the area of the parent browser window.                           */
/*                                                                      */
/*  2. A window may have an invisible margin surrounding the visible    */
/*     border edge. The tree of Tile objects representing the layout    */
/*     uses coordinates based on the visible area of tiled windows.     */
/*     The margin width is dependent on the operating system            */
/*     (e.g. 7 pixels for Windows 10, 0 pixels for Linux & Mac OS X).   */
/*                                                                      */
/*  3. A window has a minimum width that is dependent on the browser,   */
/*     operating system and the number of tabs in the window.           */
/*                                                                      */
/*  4. A window has a minimum height that is dependent on the browser,  */
/*     operating system and the number of toolbars in the window.       */
/*                                                                      */
/*  5. The minimum width and height for tiled windows are determined    */
/*     before creation of the tiles in the layout.                      */
/*                                                                      */
/*  6. The synchronize scroll feature is available only in tiled        */
/*     windows displayed in the tiled view.                             */
/*                                                                      */
/*  7. The open link in next/previous tab feature is available only     */
/*     in tiled windows displayed in the tiled view.                    */
/*                                                                      */
/*  8. Operations involving multiple windows are performed in a way     */
/*     that ensures a single completion point. This is done using       */
/*     recursive functions to traverse the layout (which also act as    */
/*     closures) and window counts.                                     */
/*                                                                      */
/*  9. Resizing adjacent tiles by dragging their edges is quite         */ 
/*     awkward to implement, since the Chrome API does not provide      */
/*     a chrome.windows.onUpdated.addListener function.                 */
/*                                                                      */
/* 10. A timer-driven 'tile event handler' function monitors changes    */
/*     in the left/top/width/height properties of the tiled windows.    */
/*     This handler has been designed to minimize the  overheads of     */
/*     polling, whilst retaining reasonable user responsiveness.        */
/*                                                                      */
/* 11. In Chrome, removing tiles using chrome.tabs.remove causes        */
/*     problems when mixed with tab moves and window updates.  For a    */
/*     short period after the callback function is called, querying     */
/*     tabs using chrome.tabs.query still returns the removed tab.      */
/*                                                                      */
/************************************************************************/

/************************************************************************/
/*                                                                      */
/* Layout String Formats                                                */
/*                                                                      */
/* Layout  Debut  Record                                                */
/*  Type    Ver   Format                                                */
/*                                                                      */
/*   all    1.0   layouttype,windowcount;                               */
/*                                                                      */
/*   all    1.0   hsplit,ratio,childcount;                              */
/*   all    1.0   vsplit,ratio,childcount;                              */
/*                                                                      */
/*    0     1.0   window,ratio;                                         */
/*    1     1.0   window,ratio,{taburi};                                */
/*    2     6.0   window,ratio,windowtype;                              */
/*    3     6.0   window,ratio,windowtype,{taburi};                     */
/*    4     6.1   window,ratio,windowtype;                              */
/*    5     6.1   window,ratio,windowtype,{taburi},scrollX,scrollY;     */
/*                                                                      */
/************************************************************************/

"use strict";

/************************************************************************/

/* Global variables */

var enableInspect = false;
var enableEventHandler = true;

var isFirefox;

var platformOS;

var windowMinWidth,windowMinHeight;

var blankPage,newtabPage;

var showSubmenu,showQuickMenu,showToolbars,setToggleAll;
var assignExist,autoCloseOpen,autoCloseUser;
var topMargin,lrbMarginNormal,lrbMarginPopup,maxAdjust;
var defaultType,defaultPage;
var openLinks;
var syncActive;

var parentWindowId = -1;
var parentWindowBlankTabId = -1;

var dialogWindowId = -1;
var dialogParentWindowId = -1;

var rootTile = null;
var windowCount = 0;
var layoutLeft,layoutTop,layoutWidth,layoutHeight;

var tiledView = false;
var syncScroll = false;
var expandedTile = null;
var lastFocusedTile = null;

var newLayoutSubmenu;

var closeTabIdList = new Array();
var autocloseTabIdList = new Array();

var intervalTimer = null;
var pollInterval = 100;  /* 100ms */
var idleCounter = -1;

var resizeHoriz,resizeBefore;

var deferTimeout = 50;  /* 50 ms */

/************************************************************************/

/* Tile object constructor */

function Tile()
{
    this.parentTile = null;
    this.type = "window";  /* window, hsplit or vsplit */
    
    /* Tile position and szie*/
    
    this.left = 0;
    this.top = 0;
    this.width = 0;
    this.height = 0;
    this.ratio = 1;
    
    /* Window tile */
    
    this.windowType = showToolbars ? "normal" : "popup";
    
    this.windowId = -1;  /* window for tile, -1 = no window */
    
    this.tabId = -1;  /* assigned tab for tile, -1 = no assigned tab */
    this.tabPos = -1;  /* position of assigned tab in parent window, -1 = no assigned tab, -2 = generated tab, -3 = tab created by Open Layout */
    
    this.scrollX = 0;  /* last scroll X */
    this.scrollY = 0;  /* last scroll Y */
    this.scrollInit = false;  /* initialize scroll position on first page load after opening layout */
    
    /* Split tile */
    
    this.childTiles = new Array();
    
    /* Temporary properties */
    
    this.tabUrl = "";  /* used when opening layout */
}

/************************************************************************/

/* Initialize on browser startup */

chrome.runtime.getPlatformInfo(
function(PlatformInfo)   
{
    platformOS = PlatformInfo.os;
    
    chrome.storage.local.set({ "environment-platformos": platformOS });
    
    isFirefox = (navigator.userAgent.indexOf("Firefox") >= 0);

    chrome.storage.local.set({ "environment-isfirefox": isFirefox });
    
    initialize();
});

function initialize()
{
    blankPage = chrome.runtime.getURL("blank.html");
    newtabPage = (isFirefox) ? "about:newtab" : "chrome://newtab";
    
    chrome.storage.local.get(null,
    function(object)
    {
        /* Initialize or migrate options */
        
        /* General options */
        
        if (!("options-showsubmenu" in object)) object["options-showsubmenu"] = true;
        if (!("options-showquickmenu" in object)) object["options-showquickmenu"] = true;
        if (!("options-showtoolbars" in object)) object["options-showtoolbars"] = true;
        if (!("options-settoggleall" in object)) object["options-settoggleall"] = false;
        
        if (!("options-assignexist" in object)) object["options-assignexist"] = true;
        if (!("options-autocloseopen" in object)) object["options-autocloseopen"] = 
            ("options-autoclose" in object) ? object["options-autoclose"] : false;  /* Version 1.0-6.5 */
        if (!("options-autocloseuser" in object)) object["options-autocloseuser"] = false;
        
        if (platformOS == "win")
        {
            if (!("options-topmargin" in object)) object["options-topmargin"] = 0;
            if (!("options-lrbmarginnormal" in object)) object["options-lrbmarginnormal"] =
                ("options-lrbmargin" in object) ? object["options-lrbmargin"] : 7;  /* Version 1.0-6.5 */
            if (!("options-lrbmarginpopup" in object)) object["options-lrbmarginpopup"] = 7;
            if (!("options-maxadjust" in object)) object["options-maxadjust"] = 1;
        }
        else  /* "linux", "mac" or other platform */
        {
            if (!("options-topmargin" in object)) object["options-topmargin"] = 0;
            if (!("options-lrbmarginnormal" in object)) object["options-lrbmarginnormal"] =
                ("options-lrbmargin" in object) ? object["options-lrbmargin"] : 0;  /* Version 1.0-6.5 */
            if (!("options-lrbmarginpopup" in object)) object["options-lrbmarginpopup"] = 0;
            if (!("options-maxadjust" in object)) object["options-maxadjust"] = 0;
        }
        
        /* Page Settings options */
        
        if (!("options-defaulttype" in object)) object["options-defaulttype"] = 0;
        if (!("options-defaultpage" in object)) object["options-defaultpage"] = 
            ("options-defaulturl" in object) ? object["options-defaulturl"] : "";  /* Version 1.0-6.5 */
        
        if (!("options-openlinks" in object)) object["options-openlinks"] = 3;
        
        if (!("options-syncactive" in object)) object["options-syncactive"] = false;
        
        if (!("layout-default" in object)) object["layout-default"] = "";
        if (!("layout-savetabs" in object)) object["layout-savetabs"] = false;
        
        if (!("reinstate-flag" in object)) object["reinstate-flag"] = false;
        
        /* Update stored options */
        
        chrome.storage.local.set(object);
        
        /* Initialize local options */
        
        showSubmenu = object["options-showsubmenu"];
        showQuickMenu = object["options-showquickmenu"];
        showToolbars = object["options-showtoolbars"];
        setToggleAll = object["options-settoggleall"];
        
        assignExist = object["options-assignexist"];
        autoCloseOpen = object["options-autocloseopen"];
        autoCloseUser = object["options-autocloseuser"];
        
        defaultType = object["options-defaulttype"];
        defaultPage = object["options-defaultpage"];
        
        openLinks = object["options-openlinks"];
        
        syncActive = object["options-syncactive"];
        
        topMargin = object["options-topmargin"];
        lrbMarginNormal = object["options-lrbmarginnormal"];
        lrbMarginPopup = object["options-lrbmarginpopup"];
        maxAdjust = object["options-maxadjust"];
        
        if (defaultType == 0) defaultPage = "";  /* new tab */
        else if (defaultType == 1) defaultPage = blankPage;
        
        if (isFirefox && (defaultPage.substr(0,6) == "about:" || defaultPage.substr(0,7) == "chrome:" ||
            defaultPage.substr(0,5) == "data:" || defaultPage.substr(0,11) == "javascript:")) defaultPage = "";
        
        /* Add context menu items */
        
        createContextMenu();
        
        /* Add listeners */
        
        addListeners();
        
        /* Reinstate window size */
        
        if (object["reinstate-flag"])
        {
            chrome.windows.getLastFocused({ },
            function(focuswin)
            {
                if (object["reinstate-state"] == "maximized")
                {
                    chrome.windows.update(focuswin.id,{ state: "maximized" });
                }
                else
                {
                    chrome.windows.update(focuswin.id,{ state: "normal",
                                                        left: object["reinstate-left"], top: object["reinstate-top"],
                                                        width: object["reinstate-width"], height: object["reinstate-height"] });
                }
                
                chrome.storage.local.set({ "reinstate-flag": false });
            });
        }
    });
}

/************************************************************************/

/* Add listeners */

function addListeners()
{
    /* Storage changed listener */
    
    chrome.storage.onChanged.addListener(
    function(changes,areaName)
    {
        chrome.storage.local.get(null,
        function(object)
        {
            showSubmenu = object["options-showsubmenu"];
            showQuickMenu = object["options-showquickmenu"];
            showToolbars = object["options-showtoolbars"];
            setToggleAll = object["options-settoggleall"];
            
            assignExist = object["options-assignexist"];
            autoCloseOpen = object["options-autocloseopen"];
            autoCloseUser = object["options-autocloseuser"];
            
            defaultType = object["options-defaulttype"];
            defaultPage = object["options-defaultpage"];
            
            openLinks = object["options-openlinks"];
            
            syncActive = object["options-syncactive"];
            
            topMargin = object["options-topmargin"];
            lrbMarginNormal = object["options-lrbmarginnormal"];
            lrbMarginPopup = object["options-lrbmarginpopup"];
            maxAdjust = object["options-maxadjust"];
            
            if (defaultType == 0) defaultPage = "";  /* new tab */
            else if (defaultType == 1) defaultPage = blankPage;
            
            if (isFirefox && (defaultPage.substr(0,6) == "about:" || defaultPage.substr(0,7) == "chrome:" ||
                defaultPage.substr(0,5) == "data:" || defaultPage.substr(0,11) == "javascript:")) defaultPage = "";
            
            if ("options-showsubmenu" in changes)
            {
                createContextMenu();
            }
            
            if ("options-showquickmenu" in changes)
            {
                if (rootTile != null && tiledView) setQuickMenuForAllTabsInTiledWindows(rootTile);
            }
        });
    });
    
    /* Browser action listener */
    
    chrome.browserAction.onClicked.addListener(
    function(tab)
    {
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            if (rootTile == null)  /* no layout */
            {
                chrome.storage.local.get(null,
                function(object)
                {
                    var layoutName,layoutString;
                    
                    windowCount = 0;
                    
                    parentWindowId = focuswin.id;
                    
                    setReinstateData(focuswin);
                    
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        layoutName = object["layout-default"];
                        layoutString = object["layout-string-" + layoutName];
                        
                        if (layoutString != null) openLayout_Reply(layoutString);
                        else
                        {
                            addTileFirst();
                            addTileOther(rootTile,"right");
                            adjustLayout();
                            newLayout(0);
                        }
                    });
                });
            }
            else if (findTileByWindowId(rootTile,focuswin.id) != null)  /* tiled window */
            {
                hideLayout(false);
            }
            else if (focuswin.id == parentWindowId)  /* parent window or other window */
            {
                if (tiledView) hideLayout(false);
                else
                {
                    chrome.windows.get(parentWindowId,{ },
                    function(parentwin)
                    {
                        setLayoutDimensions(parentwin,
                        function()
                        {
                            adjustLayout();
                            showLayout();
                        });
                    });
                }
            }
        });
    });
    
    /* Keyboard command listener */
    
    chrome.commands.onCommand.addListener(
    function(command)
    {
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            switch (command)
            {
                case "new-close-layout":
                    if (rootTile == null)
                    {
                        chrome.storage.local.get(null,
                        function(object)
                        {
                            var layoutName,layoutString;
                            
                            windowCount = 0;
                            
                            parentWindowId = focuswin.id;
                            
                            setReinstateData(focuswin);
                            
                            setLayoutDimensions(focuswin,
                            function()
                            {
                                layoutName = object["layout-default"];
                                layoutString = object["layout-string-" + layoutName];
                                
                                if (layoutString != null) openLayout_Reply(layoutString);
                                else
                                {
                                    addTileFirst();
                                    addTileOther(rootTile,"right");
                                    adjustLayout();
                                    newLayout(0);
                                }
                            });
                        });
                    }
                    else
                    {
                        if (tiledView) hideLayout(true);  /* hide then close */
                        else closeLayout();
                    }
                    break;
                    
                case "hide-show-layout":
                    if (rootTile != null)
                    {
                        if (tiledView) hideLayout(false);
                        else
                        {
                            chrome.windows.get(parentWindowId,{ },
                            function(parentwin)
                            {
                                setLayoutDimensions(parentwin,
                                function()
                                {
                                    adjustLayout();
                                    showLayout();
                                });
                            });
                        }
                    }
                    break;
                    
                case "refresh-layout":
                    if (rootTile != null && tiledView)
                    {
                        refreshLayout();
                    }
                    break;
                    
                case "toggle-toolbars":
                    if (rootTile != null && tiledView)
                    {
                        if (setToggleAll) toggleAllToolbars();
                        else toggleToolbars();
                    }
                    break;
            }
        });
    });
    
    /* Context menu listener */
    
    chrome.contextMenus.onClicked.addListener(
    function(info,tab)
    {
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            if (info.menuItemId.substr(-2) == "-b" || info.menuItemId.substr(-2) == "-c")
                info.menuItemId = info.menuItemId.substr(0,info.menuItemId.length-2);
                
            switch (info.menuItemId)
            {
                case "new-close-layout":
                
                    if (tiledView) hideLayout(true);  /* hide then close */
                    else closeLayout();
                    break;
                    
                case "new-layout-default":
                    chrome.storage.local.get(null,
                    function(object)
                    {
                        var layoutName,layoutString;
                        
                        windowCount = 0;
                        parentWindowId = focuswin.id;
                        setReinstateData(focuswin);
                        setLayoutDimensions(focuswin,
                        function()
                        {
                            layoutName = object["layout-default"];
                            layoutString = object["layout-string-" + layoutName];
                            
                            if (layoutString != null) openLayout_Reply(layoutString);
                            else
                            {
                                addTileFirst();
                                addTileOther(rootTile,"right");
                                adjustLayout();
                                newLayout(0);
                            }
                        });
                    });
                    break;
                    
                case "new-layout-2vert":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"right");
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-2horiz":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"below");
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-3vert":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"right");
                        addTileOther(rootTile.childTiles[1],"right");
                        equalizeTile(rootTile,false);
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-3horiz":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"below");
                        addTileOther(rootTile.childTiles[1],"below");
                        equalizeTile(rootTile,false);
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-4vertgrid":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"below");
                        addTileOther(rootTile,"right");
                        addTileOther(rootTile.childTiles[1],"below");
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-4horizgrid":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"right");
                        addTileOther(rootTile,"below");
                        addTileOther(rootTile.childTiles[1],"right");
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-6vertgrid":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"below");
                        addTileOther(rootTile,"right");
                        addTileOther(rootTile.childTiles[1],"below");
                        addTileOther(rootTile.childTiles[1],"right");
                        addTileOther(rootTile.childTiles[2],"below");
                        equalizeTile(rootTile,true);
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "new-layout-6horizgrid":
                    windowCount = 0;
                    parentWindowId = focuswin.id;
                    setReinstateData(focuswin);
                    setLayoutDimensions(focuswin,
                    function()
                    {
                        addTileFirst();
                        addTileOther(rootTile,"right");
                        addTileOther(rootTile.childTiles[1],"right");
                        addTileOther(rootTile,"below");
                        addTileOther(rootTile.childTiles[1],"right");
                        addTileOther(rootTile.childTiles[1].childTiles[1],"right");
                        equalizeTile(rootTile,true);
                        adjustLayout();
                        newLayout(0);
                    });
                    break;
                    
                case "hide-show-layout":
                    if (tiledView) hideLayout(false);
                    else
                    {
                        chrome.windows.get(parentWindowId,{ },
                        function(parentwin)
                        {
                            setLayoutDimensions(parentwin,
                            function()
                            {
                                adjustLayout();
                                showLayout();
                            });
                        });
                    }
                    break;
                    
                case "refresh-layout":
                    refreshLayout();
                    break;
                    
                case "toggle-toolbars":
                    toggleToolbars();
                    break;
                    
                case "toggle-alltoolbars":
                    toggleAllToolbars();
                    break;
                    
                case "add-tile-above":
                    addTile("above");
                    break;
                    
                case "add-tile-left":
                    addTile("left");
                    break;
                    
                case "add-tile-right":
                    addTile("right");
                    break;
                    
                case "add-tile-below":
                    addTile("below");
                    break;
                    
                case "remove-tile":
                    removeTile();
                    break;
                    
                case "expand-contract-tile":
                    if (findTileByWindowId(rootTile,focuswin.id) == expandedTile) contractTile();
                    else expandTile();
                    break;
                    
                case "inspect-tile-this":
                case "inspect-tile-before":
                case "inspect-tile-after":
                case "inspect-tile-parent":
                case "inspect-tile-gparent":
                case "inspect-tile-ggparent":
                    inspectTile(info.menuItemId.substr(13));
                    break;
                    
                case "open-save-layout":
                    if (rootTile == null)
                    {
                        setReinstateData(focuswin);
                        openLayout();
                    }
                    else
                    {
                        setReinstateData(focuswin);
                        saveLayout();
                    }
                    break;
                    
                case "delete-layout":
                    setReinstateData(focuswin);
                    deleteLayout();
                    break;
                    
                case "default-layout":
                    setReinstateData(focuswin);
                    defaultLayout();
                    break;
                    
                case "sync-scroll":
                    syncScroll = !syncScroll;
                    setQuickMenuForAllTabsInTiledWindows(rootTile);
                    chrome.contextMenus.update("sync-scroll-b",{ checked: syncScroll });
                    chrome.contextMenus.update("sync-scroll-c",{ checked: syncScroll });
                    break;
            }
        });
    });
    
    /* Tab event listeners */
    
    /* Firefox - chrome.tabs.onUpdated listener causes tabId's to change with Firefox 57+ */
    
    // chrome.tabs.onUpdated.addListener(  /* existing or new tab loaded */
    // function(tabId,changeInfo,tab)
    // {
    // });
    
    /* Firefox - workaround is to use chrome.webNavigation.onCompleted listener instead */

    chrome.webNavigation.onCompleted.addListener(  /* page loaded */
    function(details)
    {
        chrome.tabs.get(details.tabId,
        function(tab)
        {
            var tabId,tile;
            
            tabId = tab.id;
            
            if (rootTile != null && (tile = findTileByWindowId(rootTile,tab.windowId)) != null)  /* tiled window */
            {
                setBadgeForAllTabsInWindow(tab.windowId,3 );
                
                chrome.tabs.sendMessage(tabId,{ type: "addListeners" },checkError);
                
                if (showQuickMenu) chrome.tabs.sendMessage(tabId,{ type: "addQuickMenu", expanded: (tile == expandedTile), syncscroll: syncScroll },checkError);
                else chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
            }
            else if (tab.windowId == parentWindowId)  /* parent window */
            {
                if (tiledView) setBadgeForAllTabsInWindow(tab.windowId,2);
                else setBadgeForAllTabsInWindow(tab.windowId,1);
                
                chrome.tabs.sendMessage(tabId,{ type: "removeListeners" },checkError);
                
                chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
            }
            else  /* other window */
            {
                setBadgeForAllTabsInWindow(tab.windowId,0);
                
                chrome.tabs.sendMessage(tabId,{ type: "removeListeners" },checkError);
                
                chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
            }
        });
    });
    
    chrome.tabs.onAttached.addListener(  /* tab moved into window */
    function(tabId,attachInfo)
    {
        var tile;
        
        if (rootTile != null && (tile = findTileByWindowId(rootTile,attachInfo.newWindowId)) != null)  /* tiled window */
        {
            setBadgeForAllTabsInWindow(attachInfo.newWindowId,3);
            
            chrome.tabs.sendMessage(tabId,{ type: "addListeners" },checkError);
            
            if (showQuickMenu) chrome.tabs.sendMessage(tabId,{ type: "addQuickMenu", expanded: (tile == expandedTile), syncscroll: syncScroll },checkError);
            else chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
        }
        else if (attachInfo.newWindowId == parentWindowId)  /* parent window */
        {
            if (tiledView) setBadgeForAllTabsInWindow(attachInfo.newWindowId,2);
            else setBadgeForAllTabsInWindow(attachInfo.newWindowId,1);
            
            chrome.tabs.sendMessage(tabId,{ type: "removeListeners" },checkError);
            
            chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
        }
        else  /* other window */
        {
            setBadgeForAllTabsInWindow(attachInfo.newWindowId,0);
            
            chrome.tabs.sendMessage(tabId,{ type: "removeListeners" },checkError);
            
            chrome.tabs.sendMessage(tabId,{ type: "removeQuickMenu" },checkError);
        }
    });
    
    /* Window event listeners */
    
    chrome.windows.onFocusChanged.addListener(  /* fired when browser window gains focus or when all browser windows lose focus */
    function(windowId)
    {
        var flag;
        
        if (windowId == chrome.windows.WINDOW_ID_NONE) return;  /* all Chrome windows have lost focus */
        
        updateContextMenuItems(windowId);
        
        if (dialogWindowId != -1) chrome.windows.update(dialogWindowId,{ focused: true });  /* keep dialog window on top */
    });
    
    chrome.windows.onRemoved.addListener(  /* window closed */
    function(windowId)
    {
        var tile;
        
        /* Event is fired when 'window' is closed as a result of the following actions: */
        /* 1. User clicking on close (x) button of 'window' or close (x) button of last tab in 'window' */
        /* 2. User dragging last tab out of 'window' and releasing in another window */
        /* 3. Program moving last tab out of tiled 'window' by calling chrome.tabs.move - as in hideLayout(), toggleToolbars() or removeTile() */
        /* 4. User dragging any tab out of window and releasing in another window causing temporary drag 'window' to close */
        /* 5. User clicking on open/save/delete/cancel button of dialog 'window' */
        /* 6. Program removing a dialog 'window' by calling chrome.windows.remove() */
        
        if (rootTile != null && (tile = findTileByWindowId(rootTile,windowId)) != null)  /* tiled window */  /* cases 1 or 2 */
        {
            stopEventHandler();
           
            setTimeout(  /* allow time for tileEventHandler() to complete */
            function()
            {
                lastFocusedTile = removeTileOther(tile);
                
                if (windowCount == 1) hideLayout(true);
                else
                {
                    adjustLayout();
                    
                    _refreshLayout();  /* will restart eventHandler */
                }
            },deferTimeout);
        }
        else if (windowId == parentWindowId)  /* parent window */  /* cases 1 or 2 */ 
        {
            parentWindowId = -1;
            parentWindowBlankTabId = -1;
        }
        else if (windowId == dialogWindowId)  /* dialog window */  /* cases 1, 5 or 6 */
        {
            dialogWindowId = -1;
            dialogParentWindowId = -1;
        }
        else ;  /* other window */  /* cases 1, 2, 3 or 4 */
        
        if (windowId == dialogParentWindowId)  /* dialog parent window */  /* cases 1, 2 or 3 */
        {
            chrome.windows.remove(dialogWindowId);
            
            dialogWindowId = -1;
            dialogParentWindowId = -1;
        }
    });
    
    /* Message received listener */
    
    chrome.runtime.onMessage.addListener(
    function(message,sender,sendResponse)
    {
        var i,tile;
        
        switch (message.type)
        {
            /* Messages from content scripts */
            
            case "pageLoad":
                
                if (rootTile != null && (tile = findTileByWindowId(rootTile,sender.tab.windowId)) != null)  /* tiled window */
                {
                    chrome.tabs.sendMessage(sender.tab.id,{ type: "addListeners" },checkError);
                    
                    if (showQuickMenu) chrome.tabs.sendMessage(sender.tab.id,{ type: "addQuickMenu", expanded: (tile == expandedTile), syncscroll: syncScroll },checkError);
                    else chrome.tabs.sendMessage(sender.tab.id,{ type: "removeQuickMenu" },checkError);
                    
                    if (tile.scrollInit) chrome.tabs.sendMessage(sender.tab.id,{ type: "initScrollPosn", scrollx: tile.scrollX, scrolly: tile.scrollY },checkError);
                    tile.scrollInit = false;
                }
                else
                {
                    chrome.tabs.sendMessage(sender.tab.id,{ type: "removeListeners" },checkError);
                    
                    chrome.tabs.sendMessage(sender.tab.id,{ type: "removeQuickMenu" },checkError);
                }
                
                break;
                
            case "closeLayout":
                
                if (tiledView) hideLayout(true);  /* hide then close */
                else closeLayout();
                
                break;
                
            case "hideLayout":
                
                hideLayout(false);
                
                break;
                
            case "refreshLayout":
                
                refreshLayout();
                
                break;
                
            case "toggleToolbars":
                
                toggleToolbars();
                
                break;
                
            case "toggleAllToolbars":
                
                toggleAllToolbars();
                
                break;
                
            case "addTile":
                
                addTile(message.direction);
                
                break;
                
            case "removeTile":
                
                removeTile();
                
                break;
                
            case "expandTile":
                
                if (findTileByWindowId(rootTile,sender.tab.windowId) == expandedTile) contractTile();
                else expandTile();
                
                break;
                
            case "syncScroll":
                
                syncScroll = !syncScroll;
                setQuickMenuForAllTabsInTiledWindows(rootTile);
                chrome.contextMenus.update("sync-scroll-b",{ checked: syncScroll });
                chrome.contextMenus.update("sync-scroll-c",{ checked: syncScroll });
                
                break;
                
            case "openInNext":
                
                if (rootTile != null && (tile = findTileByWindowId(rootTile,sender.tab.windowId)) != null)  /* tiled window */
                {
                    if (tiledView)  /* tiled window */
                    {
                        for (i = 0; i < tile.parentTile.childTiles.length; i++) if (tile.parentTile.childTiles[i] == tile) break;
                        
                        if (message.prev)  /* open in previous tile */
                        {
                            if (i == 0) i = tile.parentTile.childTiles.length-1; else i--;
                        }
                        else  /* open in next tile */
                        {
                            if (i == tile.parentTile.childTiles.length-1) i = 0; else i++;
                        }
                        
                        chrome.tabs.query({ windowId: tile.parentTile.childTiles[i].windowId, active: true },
                        function(tabs)
                        {
                            chrome.tabs.update(tabs[0].id, { url: message.href });
                        });
                    }
                }
                
                break;
                
            case "onScroll":
                
                chrome.windows.getLastFocused({ },
                function(focuswin)
                {
                    var i,tile,deltaX,deltaY;
                    
                    if (rootTile != null && (tile = findTileByWindowId(rootTile,sender.tab.windowId)) != null)  /* tiled window */
                    {
                        if (tiledView)  /* tiled view */
                        {
                            deltaX = message.scrollx-tile.scrollX;
                            deltaY = message.scrolly-tile.scrollY;
                            
                            if (syncScroll && (!syncActive || sender.tab.windowId == focuswin.id))
                            {
                                for (i = 0; i < tile.parentTile.childTiles.length; i++)
                                {
                                    if (tile.parentTile.childTiles[i] != tile && tile.parentTile.childTiles[i].type == "window")
                                    {
                                        tile.parentTile.childTiles[i].scrollX += deltaX;
                                        tile.parentTile.childTiles[i].scrollY += deltaY;
                                        
                                        chrome.tabs.sendMessage(tile.parentTile.childTiles[i].tabId,{ type: "setScrollPosn", deltax: deltaX, deltay: deltaY },checkError);
                                    }
                                }
                            }
                            
                            tile.scrollX += deltaX;
                            tile.scrollY += deltaY;
                        }
                    }
                });
                
                break;
                
            /* Messages from scripts in dialog pages */
            
            case "openLayout_Reply":
                
                chrome.windows.update(dialogParentWindowId,{ focused: true },
                function(focuswin)
                {
                    if (message.result)
                    {
                        chrome.storage.local.get(null,
                        function(object)
                        {
                            var layoutName,layoutString;
                            
                            windowCount = 0;
                            
                            parentWindowId = focuswin.id;
                            
                            setReinstateData(focuswin);
                            
                            setLayoutDimensions(focuswin,
                            function()
                            {
                                layoutString = object["layout-string-" + message.layoutName];
                                
                                window.setTimeout(function() { openLayout_Reply(layoutString); },0);  /* break execution otherwise Firefox fails */
                            });
                        });
                    }
                    
                    chrome.windows.remove(dialogWindowId);
                    
                    dialogWindowId = -1;
                    dialogParentWindowId = -1;
                });
                
                break;
                
            case "saveLayout_Reply":
                
                if (message.result) saveLayout_Reply(message.layoutName,message.saveTabs);
                
                chrome.windows.remove(dialogWindowId);
                
                dialogWindowId = -1;
                dialogParentWindowId = -1;
                
                break;
                
            case "deleteLayout_Reply":
                
                if (message.result) deleteLayout_Reply(message.layoutName);
                
                chrome.windows.remove(dialogWindowId);
                
                dialogWindowId = -1;
                dialogParentWindowId = -1;
                
                break;
                
            case "defaultLayout_Reply":
                
                if (message.result) defaultLayout_Reply(message.layoutName);
                
                chrome.windows.remove(dialogWindowId);
                
                dialogWindowId = -1;
                dialogParentWindowId = -1;
                
                break;
        }
    });
}

/************************************************************************/

/* Add tile functions */

function addTileFirst()
{
    rootTile = new Tile();
    rootTile.left = layoutLeft;
    rootTile.top = layoutTop;
    rootTile.width = layoutWidth;
    rootTile.height = layoutHeight;
    
    windowCount++;
    
    return rootTile;
}

function addTileOther(tile,direction)
{
    var i,pc,newTile,modTile;
    
    /* If 'tile.type' is 'window', then 'direction' can be any direction */
    /* If 'tile.type' is 'vsplit', then 'direction' must be 'left' or 'right' */
    /* If 'tile.type' is 'hsplit', then 'direction' must be 'above' or 'below' */
     
    if (tile.parentTile == null && tile.type != "window")
    {
        rootTile = new Tile();
        rootTile.type = (direction == "left" || direction == "right") ? "vsplit" : "hsplit";
        rootTile.left = layoutLeft;
        rootTile.top = layoutTop;
        rootTile.width = layoutWidth;
        rootTile.height = layoutHeight;
        rootTile.childTiles[0] = tile;
        tile.parentTile = rootTile;
    }
    
    if (tile.parentTile != null &&
        ( (tile.parentTile.type == "vsplit" && (direction == "left" || direction == "right")) ||
          (tile.parentTile.type == "hsplit" && (direction == "above" || direction == "below")) ) )   /* add sibling window tile to parent tile */
    {
        for (pc = 0; pc < tile.parentTile.childTiles.length; pc++) if (tile.parentTile.childTiles[pc] == tile) break;
        tile = tile.parentTile;
        
        switch (direction)
        {
            case "left":
                for (i = tile.childTiles.length-1; i >= pc; i--)
                    tile.childTiles[i+1] = tile.childTiles[i];
                tile.childTiles[pc] = new Tile();
                tile.childTiles[pc].parentTile = tile;
                tile.childTiles[pc].width = Math.floor((tile.childTiles[pc+1].width)/2);
                tile.childTiles[pc].height = tile.height;
                tile.childTiles[pc].left = tile.childTiles[pc+1].left;
                tile.childTiles[pc].top = tile.top;
                tile.childTiles[pc+1].left = tile.childTiles[pc+1].left+Math.floor((tile.childTiles[pc+1].width)/2);
                tile.childTiles[pc+1].width = tile.childTiles[pc+1].width-Math.floor((tile.childTiles[pc+1].width)/2);
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].width/tile.width;
                if (tile.childTiles[pc+1].type != "window")
                    resizeTile(tile.childTiles[pc+1],tile.childTiles[pc+1].width,tile.height,tile.childTiles[pc+1].left,tile.top);
                break;
                
            case "right":
                for (i = tile.childTiles.length-1; i > pc; i--)
                    tile.childTiles[i+1] = tile.childTiles[i];
                tile.childTiles[pc+1] = new Tile();
                tile.childTiles[pc+1].parentTile = tile;
                tile.childTiles[pc+1].width = tile.childTiles[pc].width-Math.floor((tile.childTiles[pc].width)/2);
                tile.childTiles[pc+1].height = tile.height;
                tile.childTiles[pc+1].left = tile.childTiles[pc].left+Math.floor((tile.childTiles[pc].width)/2);
                tile.childTiles[pc+1].top = tile.top;
                tile.childTiles[pc].width = Math.floor((tile.childTiles[pc].width)/2);
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].width/tile.width;
                if (tile.childTiles[pc].type != "window")
                    resizeTile(tile.childTiles[pc],tile.childTiles[pc].width,tile.height,tile.childTiles[pc].left,tile.top);
                break;
                
            case "above":
                for (i = tile.childTiles.length-1; i >= pc; i--)
                    tile.childTiles[i+1] = tile.childTiles[i];
                tile.childTiles[pc] = new Tile();
                tile.childTiles[pc].parentTile = tile;
                tile.childTiles[pc].width = tile.width;
                tile.childTiles[pc].height = Math.floor((tile.childTiles[pc+1].height)/2);
                tile.childTiles[pc].left = tile.left;
                tile.childTiles[pc].top = tile.childTiles[pc+1].top;
                tile.childTiles[pc+1].top = tile.childTiles[pc+1].top+Math.floor((tile.childTiles[pc+1].height)/2);
                tile.childTiles[pc+1].height = tile.childTiles[pc+1].height-Math.floor((tile.childTiles[pc+1].height)/2);
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].height/tile.height;
                if (tile.childTiles[pc+1].type != "window")
                    resizeTile(tile.childTiles[pc+1],tile.width,tile.childTiles[pc+1].height,tile.left,tile.childTiles[pc+1].top);
                break;
                
            case "below":
                for (i = tile.childTiles.length-1; i > pc; i--)
                    tile.childTiles[i+1] = tile.childTiles[i];
                tile.childTiles[pc+1] = new Tile();
                tile.childTiles[pc+1].parentTile = tile;
                tile.childTiles[pc+1].width = tile.width;
                tile.childTiles[pc+1].height = tile.childTiles[pc].height-Math.floor((tile.childTiles[pc].height)/2);
                tile.childTiles[pc+1].left = tile.left;
                tile.childTiles[pc+1].top = tile.childTiles[pc].top+Math.floor((tile.childTiles[pc].height)/2);
                tile.childTiles[pc].height = Math.floor((tile.childTiles[pc].height)/2);
                for (i = 0; i < tile.childTiles.length; i++)
                    tile.childTiles[i].ratio = tile.childTiles[i].height/tile.height;
                if (tile.childTiles[pc].type != "window")
                    resizeTile(tile.childTiles[pc],tile.width,tile.childTiles[pc].height,tile.left,tile.childTiles[pc].top);
                break;
        }
        
        newTile = (direction == "left" || direction == "above") ? tile.childTiles[pc] : tile.childTiles[pc+1];
        modTile = (direction == "left" || direction == "above") ? tile.childTiles[pc+1] : tile.childTiles[pc];
    }
    else /* split this window tile into two child window tiles */
    {
        tile.childTiles[0] = new Tile();
        tile.childTiles[0].parentTile = tile;
        tile.childTiles[0].width = (direction == "left" || direction == "right") ? Math.floor(tile.width/2) : tile.width;
        tile.childTiles[0].height = (direction == "left" || direction == "right") ? tile.height : Math.floor(tile.height/2);
        tile.childTiles[0].left = tile.left;
        tile.childTiles[0].top = tile.top;
        tile.childTiles[0].ratio = 1/2;
        tile.childTiles[0].windowType = (direction == "left" || direction == "above") ? (showToolbars ? "normal" : "popup") : tile.windowType;
        tile.childTiles[0].windowId = (direction == "left" || direction == "above") ? -1 : tile.windowId;
        tile.childTiles[0].tabId = (direction == "left" || direction == "above") ? -1 : tile.tabId;
        tile.childTiles[0].tabPos = (direction == "left" || direction == "above") ? -1 : tile.tabPos;
        tile.childTiles[1] = new Tile();
        tile.childTiles[1].parentTile = tile;
        tile.childTiles[1].width = (direction == "left" || direction == "right") ? tile.width-tile.childTiles[0].width : tile.width;
        tile.childTiles[1].height = (direction == "left" || direction == "right") ? tile.height : tile.height-tile.childTiles[0].height;
        tile.childTiles[1].left = (direction == "left" || direction == "right") ? tile.left+tile.childTiles[0].width : tile.left;
        tile.childTiles[1].top = (direction == "left" || direction == "right") ? tile.top : tile.top+tile.childTiles[0].height;
        tile.childTiles[1].ratio = 1/2;
        tile.childTiles[1].windowType = (direction == "left" || direction == "above") ? tile.windowType : (showToolbars ? "normal" : "popup");
        tile.childTiles[1].windowId = (direction == "left" || direction == "above") ? tile.windowId : -1;
        tile.childTiles[1].tabId = (direction == "left" || direction == "above") ? tile.tabId : -1;
        tile.childTiles[1].tabPos = (direction == "left" || direction == "above") ? tile.tabPos : -1;
        tile.type = (direction == "left" || direction == "right") ? "vsplit" : "hsplit";
        tile.tabId = -1;
        tile.tabPos = -1;
        
        newTile = (direction == "left" || direction == "above") ? tile.childTiles[0] : tile.childTiles[1];
        modTile = (direction == "left" || direction == "above") ? tile.childTiles[1] : tile.childTiles[0];
    }
    
    windowCount++;
    
    return { newTile: newTile, modTile: modTile };
}

/************************************************************************/
    
/* Remove tile function */

function removeTileOther(tile)
{
    var i,num,pc,gpc,gptile,ptile,stile,ftile;
    
    ptile = tile.parentTile;
    for (pc = 0; pc < ptile.childTiles.length; pc++) if (ptile.childTiles[pc] == tile) break;
    if (pc == 0) stile = ptile.childTiles[1]; else stile = ptile.childTiles[pc-1];  /* sibling tile */
    
    if (ptile.childTiles.length == 2)  /* exactly 2 children */
    {
        if (stile.type == "window")  /* 2 window tiles */
        {
            ptile.type = "window";
            ptile.windowType = stile.windowType;
            ptile.windowId = stile.windowId;
            ptile.tabId = stile.tabId;
            ptile.tabPos = stile.tabPos;
            
            ptile.childTiles.length = 0;
            
            windowCount--;
            
            return stile;
        }
        else  /* 1 window tile and 1 hsplit/vsplit tile */
        {
            for (ftile = stile.childTiles[0]; ftile.type != "window"; ftile = tile.childTiles[0]);
            
            if (ptile.parentTile == null) /* at root */
            {
                if (ptile.type == "hsplit")  /* horizontal split */
                {
                    stile.height = ptile.height;
                    stile.top = ptile.top;
                    stile.ratio = 1;
                    stile.parentTile = null;
                }
                else  /* vertical split */
                {
                    stile.width = ptile.width;
                    stile.left = ptile.left;
                    stile.ratio = 1;
                    stile.parentTile = null;
                }
                
                ptile.childTiles.length = 0;
                
                resizeTile(stile,stile.width,stile.height,stile.left,stile.top);
                
                rootTile = stile;
            }
            else  /* not at root */
            {
                num = stile.childTiles.length;
                gptile = ptile.parentTile;
                for (gpc = 0; gpc < gptile.childTiles.length; gpc++) if (gptile.childTiles[gpc] == ptile) break;
                for (i = gptile.childTiles.length-1; i > gpc; i--)
                {
                    gptile.childTiles[i+num-1] = gptile.childTiles[i];
                }
                for (i = 0; i < num; i++) gptile.childTiles[gpc+i] = stile.childTiles[i];
                if (gptile.type == "hsplit")  /* horizontal split */
                {
                    for (i = 0; i < num; i++)
                    {
                        gptile.childTiles[gpc+i].parentTile = gptile;
                        gptile.childTiles[gpc+i].width = gptile.width;
                        gptile.childTiles[gpc+i].left = gptile.left;
                    }
                    for (i = 0; i < gptile.childTiles.length; i++)
                      gptile.childTiles[i].ratio = gptile.childTiles[i].height/gptile.height;
                }
                else  /* vertical split */
                {
                    for (i = 0; i < num; i++)
                    {
                        gptile.childTiles[gpc+i].parentTile = gptile;
                        gptile.childTiles[gpc+i].height = gptile.height;
                        gptile.childTiles[gpc+i].top = gptile.top;
                    }
                    for (i = 0; i < gptile.childTiles.length; i++)
                        gptile.childTiles[i].ratio = gptile.childTiles[i].width/gptile.width;
                }
                
                stile.childTiles.length = 0;
                ptile.childTiles.length = 0;
                
                resizeTile(gptile,gptile.width,gptile.height,gptile.left,gptile.top);
            }
            
            windowCount--;
            
            return ftile;
        }
    }
    else  /* more than 2 children */
    {
        for (ftile = stile; ftile.type != "window"; ftile = tile.childTiles[0]);
        
        if (ptile.type == "hsplit")  /* horizontal split */
        {
            stile.height += tile.height;
            if (pc == 0) stile.top = tile.top;
            for (i = 0; i < ptile.childTiles.length; i++) if (i != pc)
                ptile.childTiles[i].ratio = ptile.childTiles[i].height/ptile.height;
        }
        else  /* vertical split */
        {
            stile.width += tile.width;
            if (pc == 0) stile.left = tile.left;
            for (i = 0; i < ptile.childTiles.length; i++) if (i != pc)
                ptile.childTiles[i].ratio = ptile.childTiles[i].width/ptile.width;
        }
        for (i = pc; i < ptile.childTiles.length-1; i++)
        {
            ptile.childTiles[i] = ptile.childTiles[i+1];
        }

        ptile.childTiles.length--;

        resizeTile(ptile,ptile.width,ptile.height,ptile.left,ptile.top);
        
        windowCount--;
        
        return ftile;
    }
}

function removeTileLast()
{
    rootTile = null;
    
    windowCount--;
}
    
/************************************************************************/

/* Resize tile functions */

function equalizeTile(tile,nested)
{
    var i;
    
    if (tile.type == "window")  /* window tile */
    {
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
        {
            tile.childTiles[i].ratio = 1/tile.childTiles.length;
            if (nested) equalizeTile(tile.childTiles[i],nested);
        }
    }
}

function resizeTile(tile,newWidth,newHeight,newLeft,newTop)
{
    var i,width,height,prewidth,preheight;
    
    if (tile.type == "window")  /* window tile */
    {
        tile.width = newWidth;
        tile.height = newHeight;
        tile.left = newLeft;
        tile.top = newTop;
    }
    else  /* split tile */
    {
        if (tile.type == "hsplit")  /* horizontal split */
        {
            preheight = 0;
            for (i = 0; i < tile.childTiles.length-1; i++)
            {
                height = Math.floor(newHeight*tile.childTiles[i].ratio);
                resizeTile(tile.childTiles[i],newWidth,height,newLeft,newTop+preheight);
                preheight += height;
            }
            resizeTile(tile.childTiles[i],newWidth,newHeight-preheight,newLeft,newTop+preheight);
        }
        else  /* vertical split */
        {
            prewidth = 0;
            for (i = 0; i < tile.childTiles.length-1; i++)
            {
                width = Math.floor(newWidth*tile.childTiles[i].ratio);
                resizeTile(tile.childTiles[i],width,newHeight,newLeft+prewidth,newTop);
                prewidth += width;
            }
            resizeTile(tile.childTiles[i],newWidth-prewidth,newHeight,newLeft+prewidth,newTop);
        }
        
        tile.width = newWidth;
        tile.height = newHeight;
        tile.left = newLeft;
        tile.top = newTop;
    }
}

/************************************************************************/

/* Adjust layout/tile functions */

function adjustLayout()
{
    /* Adjust layout to take account of window minimum width - calculate tile ratios and layout size */
    
    adjustTile(rootTile,layoutWidth,layoutHeight);
    
    /* Remember layout size */
    
    layoutWidth = rootTile.width;
    layoutHeight = rootTile.height;
    
    /* Resize layout - apply tile ratios to calculate tile sizes */
    
    resizeTile(rootTile,layoutWidth,layoutHeight,layoutLeft,layoutTop);
}

function adjustTile(tile,newWidth,newHeight)  /* must call resizeTile() after calling adjustTile() */
{
    var i,width,height,prewidth,preheight,sumwidth,sumheight,maxwidth,maxheight;
    
    if (tile.type == "window")  /* window tile */
    {
        tile.width = newWidth;
        tile.height = newHeight;
        
        if (tile.width < windowMinWidth)  tile.width = windowMinWidth;
        if (tile.height < windowMinHeight)  tile.height = windowMinHeight;
        
        // console.log("window  -  W:" + newWidth + " H:" + newHeight + "  -  W:" + tile.width + " H:" + tile.height );
    }
    else  /* split tile */
    {
        if (tile.type == "hsplit")  /* horizontal split */
        {
            preheight = sumheight = maxwidth = 0;
            for (i = 0; i < tile.childTiles.length-1; i++)
            {
                height = Math.floor(newHeight*tile.childTiles[i].ratio);
                adjustTile(tile.childTiles[i],newWidth,height);
                preheight += height;
                sumheight += tile.childTiles[i].height;
                if (tile.childTiles[i].width > maxwidth) maxwidth = tile.childTiles[i].width;
            }
            adjustTile(tile.childTiles[i],newWidth,newHeight-preheight);
            sumheight += tile.childTiles[i].height;
            if (tile.childTiles[i].width > maxwidth) maxwidth = tile.childTiles[i].width;
            
            for (i = 0; i < tile.childTiles.length; i++)
            {
                tile.childTiles[i].ratio = tile.childTiles[i].height/sumheight;
                tile.childTiles[i].width = maxwidth;
            }
            
            tile.width = maxwidth;
            tile.height = sumheight;
            
            // console.log("hsplit  -  W:" + newWidth + " H:" + newHeight + "  -  W:" + tile.width + " H:" + tile.height );
        }
        else  /* vertical split */
        {
            prewidth = sumwidth = maxheight = 0;
            for (i = 0; i < tile.childTiles.length-1; i++)
            {
                width = Math.floor(newWidth*tile.childTiles[i].ratio);
                adjustTile(tile.childTiles[i],width,newHeight);
                prewidth += width;
                sumwidth += tile.childTiles[i].width;
                if (tile.childTiles[i].height > maxheight) maxheight = tile.childTiles[i].height;
            }
            adjustTile(tile.childTiles[i],newWidth-prewidth,newHeight);
            sumwidth += tile.childTiles[i].width;
            if (tile.childTiles[i].height > maxheight) maxheight = tile.childTiles[i].height;
            
            for (i = 0; i < tile.childTiles.length; i++)
            {
                tile.childTiles[i].ratio = tile.childTiles[i].width/sumwidth;
                tile.childTiles[i].height = maxheight;
            }
            
            tile.width = sumwidth;
            tile.height = maxheight;
            
            // console.log("vsplit  -  W:" + newWidth + " H:" + newHeight + "  -  W:" + tile.width + " H:" + tile.height );
        }
    }
}

/************************************************************************/

/* Utility functions */

function setReinstateData(parentwin)
{
    chrome.storage.local.set({ "reinstate-flag": true, "reinstate-state": parentwin.state,
                               "reinstate-left": parentwin.left, "reinstate-top": parentwin.top,
                               "reinstate-width": parentwin.width, "reinstate-height": parentwin.height });
}

function setLayoutDimensions(parentwin,callback)
{
    layoutLeft = parentwin.left+lrbMarginNormal;
    layoutTop = parentwin.top+topMargin;
    layoutWidth = parentwin.width-lrbMarginNormal*2;
    layoutHeight = parentwin.height-topMargin-lrbMarginNormal;
    
    if (parentwin.state == "maximized")
    {
        layoutLeft += maxAdjust;
        layoutTop += lrbMarginNormal+maxAdjust;
        layoutWidth -= maxAdjust*2;
        layoutHeight -= lrbMarginNormal+maxAdjust*2;
    }
    
    /* Convert maximized window to normal window */
    
    // chrome.windows.update(parentwin.id,{ left: getParentWindowLeft(), top: getParentWindowTop(),
                                         // width: getParentWindowWidth(), height: getParentWindowHeight() },
    // function(parentwin)
    // {
        callback();
    // });
}

function findTileByWindowId(tile,windowId)
{
    var i,value;
    
    if (tile.type == "window")  /* window tile */
    {
        if (tile.windowId == windowId) return tile;
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
            if ((value = findTileByWindowId(tile.childTiles[i],windowId)) != null) return value;
    }
    
    return null;
}

function findTileByTabId(tile,tabId)
{
    var i,value;
    
    if (tile.type == "window")  /* window tile */
    {
        if (tile.tabId == tabId) return tile;
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
            if ((value = findTileByTabId(tile.childTiles[i],tabId)) != null) return value;
    }
    
    return null;
}

function getParentWindowLeft()
{
    return layoutLeft-lrbMarginNormal;
}

function getParentWindowTop()
{
    return layoutTop-topMargin;
}

function getParentWindowWidth()
{
    return layoutWidth+lrbMarginNormal*2;
}

function getParentWindowHeight()
{
    return layoutHeight+topMargin+lrbMarginNormal;
}

function getTiledWindowLeft(tile)
{
    return (tile.windowType == "normal") ? tile.left-lrbMarginNormal : tile.left-lrbMarginPopup;
}

function getTiledWindowTop(tile)
{
    return tile.top-topMargin;
}

function getTiledWindowWidth(tile)
{
    return (tile.windowType == "normal") ? tile.width+lrbMarginNormal*2 : tile.width+lrbMarginPopup*2;
}

function getTiledWindowHeight(tile)
{
    return (tile.windowType == "normal") ? tile.height+topMargin+lrbMarginNormal : tile.height+topMargin+lrbMarginPopup;
}

function getTileMinWidth(tile)
{
    var i,width,maxwidth,sumwidth;
    
    if (tile.type == "window")  /* window tile */
    {
        return windowMinWidth;
    }
    else if (tile.type == "vsplit") /* splitter tile */
    {
        sumwidth = 0;
        
        for (i = 0; i < tile.childTiles.length; i++)
        {
            sumwidth += getTileMinWidth(tile.childTiles[i]);
        }
        
        return sumwidth;
    }
    else if (tile.type == "hsplit") /* splitter tile */
    {
        maxwidth = 0;
        
        for (i = 0; i < tile.childTiles.length; i++)
        {
            width = getTileMinWidth(tile.childTiles[i]);
            
            if (width > maxwidth) maxwidth = width;
        }
        
        return maxwidth;
    }
}

function getTileMinHeight(tile)
{
    var i,height,maxheight,sumheight;
    
    if (tile.type == "window")  /* window tile */
    {
        return windowMinHeight;
    }
    else if (tile.type == "hsplit") /* splitter tile */
    {
        sumheight = 0;
        
        for (i = 0; i < tile.childTiles.length; i++)
        {
            sumheight += getTileMinHeight(tile.childTiles[i]);
        }
        
        return sumheight;
    }
    else if (tile.type == "vsplit") /* splitter tile */
    {
        maxheight = 0;
        
        for (i = 0; i < tile.childTiles.length; i++)
        {
            height = getTileMinHeight(tile.childTiles[i]);
            
            if (height > maxheight) maxheight = height;
        }
        
        return maxheight;
    }
}

function setBadgeForAllTabsInWindow(windowId,type)
{
    chrome.tabs.query({ windowId: windowId },
    function(tabs)
    {
        var pos,path,title,text,color;
        
        if (type == 0)
        {
            path = "icon16.png";
            title = "Tile Tabs WE";
            text = "";
            color = "#000000";
        }
        else if (type == 1)
        {
            path = "icon16.png";
            title = "Tile Tabs WE - Parent Window";
            text = "0";
            color = "#5070E0";
        }
        else if (type == 2)
        {
            path = "icon16.png";
            title = "Tile Tabs WE - Parent Window";
            text = "" + windowCount;
            color = "#5070E0";
        }
        else if (type == 3)
        {
            path = "icon16.png";
            title = "Tile Tabs WE - Tiled Window";
            text = "T";
            color = "#5070E0";
        }
        
        for (pos = 0; pos < tabs.length; pos++)
        {
            chrome.tabs.get(tabs[pos].id,
            function(tab)
            {
                if (chrome.runtime.lastError == null)  /* in case tab has been closed */
                {
                    chrome.browserAction.setIcon({ tabId: tab.id, path: path});
                    chrome.browserAction.setTitle({ tabId: tab.id, title: title });
                    chrome.browserAction.setBadgeText({ tabId: tab.id, text: text });
                    chrome.browserAction.setBadgeBackgroundColor({ tabId: tab.id, color: color });
                }
            });
        }
    });
}

function setListenersForAllTabsInWindow(windowId,enable)
{
    chrome.tabs.query({ windowId: windowId },
    function(tabs)
    {
        var pos;
        
        for (pos = 0; pos < tabs.length; pos++)
        {
            if (enable) chrome.tabs.sendMessage(tabs[pos].id,{ type: "addListeners" },checkError);
            else chrome.tabs.sendMessage(tabs[pos].id,{ type: "removeListeners" },checkError);
        }
    });
}

function setQuickMenuForAllTabsInWindow(windowId,enable)
{
    chrome.tabs.query({ windowId: windowId },
    function(tabs)
    {
        var expanded,pos;
        
        expanded = (rootTile != null && findTileByWindowId(rootTile,windowId) == expandedTile);
        
        for (pos = 0; pos < tabs.length; pos++)
        {
            if (enable) chrome.tabs.sendMessage(tabs[pos].id,{ type: "addQuickMenu", expanded: expanded, syncscroll: syncScroll },checkError);
            else chrome.tabs.sendMessage(tabs[pos].id,{ type: "removeQuickMenu" },checkError);
        }
    });
}

function setQuickMenuForAllTabsInTiledWindows(tile)
{
    var i;
    
    if (tile.type == "window")
    {
        setQuickMenuForAllTabsInWindow(tile.windowId,showQuickMenu);
    }
    else  /* split tile */
    {
        for (i = 0; i < tile.childTiles.length; i++)
            setQuickMenuForAllTabsInTiledWindows(tile.childTiles[i]);
    }
}

function createContextMenu()
{
    var aTypes,bTypes,cTypes;
    
    chrome.contextMenus.removeAll();
    
    if (isFirefox)
    {
        aTypes = showSubmenu ? [ "all" ] : [ "browser_action" ];
        bTypes = [ "browser_action" ];
        cTypes = showSubmenu ? [ "audio", "editable", "frame", "image", "link", "page", "password", "selection", "video" ] : [];
    }
    else
    {
        aTypes = showSubmenu ? [ "all" ] : [ "browser_action" ];
        bTypes = [ "browser_action" ];
        cTypes = showSubmenu ? [ "audio", "editable", "frame", "image", "link", "page", "selection", "video" ] : [];
    }
    
    /* All menus */
    
    chrome.contextMenus.create({ id: "new-close-layout", title: "New Layout", contexts: aTypes, enabled: true });
    chrome.contextMenus.create({ id: "hide-show-layout", title: "Hide Layout", contexts: aTypes, enabled: false });
    chrome.contextMenus.create({ id: "refresh-layout", title: "Refresh Layout", contexts: aTypes, enabled: false });
    chrome.contextMenus.create({ id: "toggle-toolbars", title: "Toggle Toolbars", contexts: aTypes, enabled: false });
    chrome.contextMenus.create({ id: "toggle-alltoolbars", title: "Toggle All Toolbars", contexts: aTypes, enabled: false });
    
    /* Button menu */
    
    chrome.contextMenus.create({ id: "more-b", title: "More...", contexts: bTypes, enabled: true });
    
    chrome.contextMenus.create({ id: "add-tile-b", parentId: "more-b", title: "Add Tile", contexts: [ "all" ], enabled: false });
    chrome.contextMenus.create({ id: "remove-tile-b", parentId: "more-b", title: "Remove Tile", contexts: [ "all" ], enabled: false });
    chrome.contextMenus.create({ id: "expand-contract-tile-b", parentId: "more-b", title: "Expand Tile", contexts: [ "all" ], enabled: false });
    if (enableInspect) chrome.contextMenus.create({ id: "inspect-tile-b", parentId: "more-b", title: "Inspect Tile", contexts: [ "all" ], enabled: true });
    
    chrome.contextMenus.create({ id: "add-tile-above-b", parentId: "add-tile-b", title: "Above", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-left-b", parentId: "add-tile-b", title: "Left", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-right-b", parentId: "add-tile-b", title: "Right", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-below-b", parentId: "add-tile-b", title: "Below", contexts: [ "all" ], enabled: true });
    
    if (enableInspect)
    {
        chrome.contextMenus.create({ id: "inspect-tile-this-b", parentId: "inspect-tile-b", title: "This", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-before-b", parentId: "inspect-tile-b", title: "Before", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-after-b", parentId: "inspect-tile-b", title: "After", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-parent-b", parentId: "inspect-tile-b", title: "Parent", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-gparent-b", parentId: "inspect-tile-b", title: "G Parent", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-ggparent-b", parentId: "inspect-tile-b", title: "GG Parent", contexts: [ "all" ], enabled: true });
    }
    
    chrome.contextMenus.create({ type: "separator", parentId: "more-b", contexts: [ "all" ], enabled: true });

    chrome.contextMenus.create({ id: "open-save-layout-b", parentId: "more-b", title: "Open Layout...", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "delete-layout-b", parentId: "more-b", title: "Delete Layout...", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "default-layout-b", parentId: "more-b", title: "Default Layout...", contexts: [ "all" ], enabled: true });
    
    chrome.contextMenus.create({ type: "separator", parentId: "more-b", contexts: [ "all" ], enabled: true });

    chrome.contextMenus.create({ id: "sync-scroll-b", parentId: "more-b", title: "Sync Scroll", type: "checkbox", contexts: [ "all" ], enabled: false, checked: false });
    
    /* Context menu */
    
    chrome.contextMenus.create({ type: "separator", contexts: cTypes, enabled: true });
    
    chrome.contextMenus.create({ id: "add-tile-c", title: "Add Tile", contexts: cTypes, enabled: false });
    chrome.contextMenus.create({ id: "remove-tile-c", title: "Remove Tile", contexts: cTypes, enabled: false });
    chrome.contextMenus.create({ id: "expand-contract-tile-c", title: "Expand Tile", contexts: cTypes, enabled: false });
    if (enableInspect) chrome.contextMenus.create({ id: "inspect-tile-c", title: "Inspect Tile", contexts: cTypes, enabled: true });
    
    chrome.contextMenus.create({ id: "add-tile-above-c", parentId: "add-tile-c", title: "Above", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-left-c", parentId: "add-tile-c", title: "Left", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-right-c", parentId: "add-tile-c", title: "Right", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "add-tile-below-c", parentId: "add-tile-c", title: "Below", contexts: [ "all" ], enabled: true });
    
    if (enableInspect)
    {
        chrome.contextMenus.create({ id: "inspect-tile-this-c", parentId: "inspect-tile-c", title: "This", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-before-c", parentId: "inspect-tile-c", title: "Before", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-after-c", parentId: "inspect-tile-c", title: "After", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-parent-c", parentId: "inspect-tile-c", title: "Parent", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-gparent-c", parentId: "inspect-tile-c", title: "G Parent", contexts: [ "all" ], enabled: true });
        chrome.contextMenus.create({ id: "inspect-tile-ggparent-c", parentId: "inspect-tile-c", title: "GG Parent", contexts: [ "all" ], enabled: true });
    }
    
    chrome.contextMenus.create({ type: "separator", contexts: cTypes, enabled: true });
    
    chrome.contextMenus.create({ id: "open-save-layout-c", title: "Open Layout...", contexts: cTypes, enabled: true });
    chrome.contextMenus.create({ id: "delete-layout-c", title: "Delete Layout...", contexts: cTypes, enabled: true });
    chrome.contextMenus.create({ id: "default-layout-c", title: "Default Layout...", contexts: cTypes, enabled: true });
    
    chrome.contextMenus.create({ type: "separator", contexts: cTypes, enabled: true });
    
    chrome.contextMenus.create({ id: "sync-scroll-c", title: "Sync Scroll", type: "checkbox", contexts: cTypes, enabled: false, checked: false });
    
    createNewLayoutSubmenu();
}

function createNewLayoutSubmenu()
{
    newLayoutSubmenu = true;
    
    chrome.contextMenus.create({ id: "new-layout-default", parentId: "new-close-layout", title: "Default Layout...", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-sep-2", type: "separator", parentId: "new-close-layout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-2vert", parentId: "new-close-layout", title: "2 Tabs - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-2horiz", parentId: "new-close-layout", title: "2 Tabs - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-sep-3", type: "separator", parentId: "new-close-layout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-3vert", parentId: "new-close-layout", title: "3 Tabs - Vertical", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-3horiz", parentId: "new-close-layout", title: "3 Tabs - Horizontal", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-sep-4", type: "separator", parentId: "new-close-layout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-4vertgrid", parentId: "new-close-layout", title: "4 Tabs - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-4horizgrid", parentId: "new-close-layout", title: "4 Tabs - Horizontal Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-sep-6", type: "separator", parentId: "new-close-layout", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-6vertgrid", parentId: "new-close-layout", title: "6 Tabs - Vertical Grid", contexts: [ "all" ], enabled: true });
    chrome.contextMenus.create({ id: "new-layout-6horizgrid", parentId: "new-close-layout", title: "6 Tabs - Horizontal Grid", contexts: [ "all" ], enabled: true });
}

function removeNewLayoutSubmenu()
{
    newLayoutSubmenu = false;
    
    chrome.contextMenus.remove("new-layout-default");
    chrome.contextMenus.remove("new-layout-sep-2");
    chrome.contextMenus.remove("new-layout-2vert");
    chrome.contextMenus.remove("new-layout-2horiz");
    chrome.contextMenus.remove("new-layout-sep-3");
    chrome.contextMenus.remove("new-layout-3vert");
    chrome.contextMenus.remove("new-layout-3horiz");
    chrome.contextMenus.remove("new-layout-sep-4");
    chrome.contextMenus.remove("new-layout-4vertgrid");
    chrome.contextMenus.remove("new-layout-4horizgrid");
    chrome.contextMenus.remove("new-layout-sep-6");
    chrome.contextMenus.remove("new-layout-6vertgrid");
    chrome.contextMenus.remove("new-layout-6horizgrid");
}

function updateContextMenuItems(windowId)
{
    var tile;
    
    if (rootTile == null)  /* no layout */
    {
        chrome.contextMenus.update("new-close-layout",{ title: "New Layout", enabled: true });
        
        chrome.contextMenus.update("hide-show-layout",{ title: "Hide Layout", enabled: false });
        
        chrome.contextMenus.update("refresh-layout",{ enabled: false });
        
        chrome.contextMenus.update("toggle-toolbars",{ enabled: false });
        
        chrome.contextMenus.update("toggle-alltoolbars",{ enabled: false });
        
        chrome.contextMenus.update("add-tile-b",{ enabled: false });
        chrome.contextMenus.update("add-tile-c",{ enabled: false });
        
        chrome.contextMenus.update("remove-tile-b",{ enabled: false });
        chrome.contextMenus.update("remove-tile-c",{ enabled: false });
        
        chrome.contextMenus.update("expand-contract-tile-b",{ title: "Expand Tile", enabled: false });
        chrome.contextMenus.update("expand-contract-tile-c",{ title: "Expand Tile", enabled: false });
        
        chrome.contextMenus.update("open-save-layout-b",{ title: "Open Layout...", enabled: true });
        chrome.contextMenus.update("open-save-layout-c",{ title: "Open Layout...", enabled: true });
        
        chrome.contextMenus.update("delete-layout-b",{ enabled: true });
        chrome.contextMenus.update("delete-layout-c",{ enabled: true });
        
        chrome.contextMenus.update("default-layout-b",{ enabled: true });
        chrome.contextMenus.update("default-layout-c",{ enabled: true });
        
        chrome.contextMenus.update("sync-scroll-b",{ enabled: false, checked: false });
        chrome.contextMenus.update("sync-scroll-c",{ enabled: false, checked: false });
        
        if (!newLayoutSubmenu) createNewLayoutSubmenu();
    }
    else if ((tile = findTileByWindowId(rootTile,windowId)) != null)  /* tiled window */
    {
        chrome.contextMenus.update("new-close-layout",{ title: "Close Layout", enabled: true });
        
        chrome.contextMenus.update("hide-show-layout",{ title: "Hide Layout", enabled: true });
        
        chrome.contextMenus.update("refresh-layout",{ enabled: true });
        
        chrome.contextMenus.update("toggle-toolbars",{ enabled: true });
        
        chrome.contextMenus.update("toggle-alltoolbars",{ enabled: true });
        
        chrome.contextMenus.update("add-tile-b",{ enabled: (windowCount >= 2) });
        chrome.contextMenus.update("add-tile-c",{ enabled: (windowCount >= 2) });
        
        chrome.contextMenus.update("remove-tile-b",{ enabled: (windowCount >= 3) });
        chrome.contextMenus.update("remove-tile-c",{ enabled: (windowCount >= 3) });
        
        if (tile == expandedTile)
        {
            chrome.contextMenus.update("expand-contract-tile-b",{ title: "Contract Tile", enabled: true });
            chrome.contextMenus.update("expand-contract-tile-c",{ title: "Contract Tile", enabled: true });
        }
        else
        {
            chrome.contextMenus.update("expand-contract-tile-b",{ title: "Expand Tile", enabled: true });
            chrome.contextMenus.update("expand-contract-tile-c",{ title: "Expand Tile", enabled: true });
        }
        
        chrome.contextMenus.update("open-save-layout-b",{ title: "Save Layout...", enabled: true });
        chrome.contextMenus.update("open-save-layout-c",{ title: "Save Layout...", enabled: true });
        
        chrome.contextMenus.update("delete-layout-b",{ enabled: true });
        chrome.contextMenus.update("delete-layout-c",{ enabled: true });
        
        chrome.contextMenus.update("default-layout-b",{ enabled: true });
        chrome.contextMenus.update("default-layout-c",{ enabled: true });
        
        chrome.contextMenus.update("sync-scroll-b",{ enabled: true, checked: syncScroll });
        chrome.contextMenus.update("sync-scroll-c",{ enabled: true, checked: syncScroll });
        
        if (newLayoutSubmenu) removeNewLayoutSubmenu();
    }
    else  /* parent window or other window */
    {
        chrome.contextMenus.update("new-close-layout",{ title: "Close Layout", enabled: true });
        
        if (tiledView) chrome.contextMenus.update("hide-show-layout",{ title: "Hide Layout", enabled: true });
        else chrome.contextMenus.update("hide-show-layout",{ title: "Show Layout", enabled: true });
        
        if (tiledView) chrome.contextMenus.update("refresh-layout",{ enabled: true });
        else chrome.contextMenus.update("refresh-layout",{ enabled: false });
        
        chrome.contextMenus.update("toggle-toolbars",{ enabled: false });
        
        chrome.contextMenus.update("toggle-alltoolbars",{ enabled: false });
        
        chrome.contextMenus.update("add-tile-b",{ enabled: false });
        chrome.contextMenus.update("add-tile-c",{ enabled: false });
        
        chrome.contextMenus.update("remove-tile-b",{ enabled: false });
        chrome.contextMenus.update("remove-tile-c",{ enabled: false });
        
        chrome.contextMenus.update("expand-contract-tile-b",{ title: "Expand Tile", enabled: false });
        chrome.contextMenus.update("expand-contract-tile-c",{ title: "Expand Tile", enabled: false });
        
        chrome.contextMenus.update("open-save-layout-b",{ title: "Save Layout...", enabled: true });
        chrome.contextMenus.update("open-save-layout-c",{ title: "Save Layout...", enabled: true });
        
        chrome.contextMenus.update("delete-layout-b",{ enabled: true });
        chrome.contextMenus.update("delete-layout-c",{ enabled: true });
        
        chrome.contextMenus.update("default-layout-b",{ enabled: true });
        chrome.contextMenus.update("default-layout-c",{ enabled: true });
        
        if (tiledView) chrome.contextMenus.update("sync-scroll-b",{ enabled: true, checked: syncScroll });
        else chrome.contextMenus.update("sync-scroll-b",{ enabled: false, checked: false });
        
        if (tiledView) chrome.contextMenus.update("sync-scroll-c",{ enabled: true, checked: syncScroll });
        else chrome.contextMenus.update("sync-scroll-c",{ enabled: false, checked: false });
        
        if (newLayoutSubmenu) removeNewLayoutSubmenu();
    }
}

/************************************************************************/

/* New layout */

function newLayout(layoutType)
{
    tiledView = true;
    expandedTile = null;
    
    chrome.windows.getLastFocused({ populate: true },
    function(focuswin)
    {
        var pos,selpos;
        
        setBadgeForAllTabsInWindow(focuswin.id,2);
        
        setListenersForAllTabsInWindow(focuswin.id,false);
        
        setQuickMenuForAllTabsInWindow(focuswin.id,false);
        
        for (selpos = 0; selpos < focuswin.tabs.length; selpos++)
            if (focuswin.tabs[selpos].active) break;  /* find selected tab */
        
        if ((layoutType == 0 || layoutType == 2 || layoutType == 4) && assignExist && windowCount >= focuswin.tabs.length)  /* prevent parent window closing */
        {
            chrome.tabs.create({ windowId: focuswin.id, url: blankPage },
            function(newtab)
            {
                parentWindowBlankTabId = newtab.id;
                
                createTiles();
            });
        }
        else createTiles();
        
        function createTiles()
        {
            var tile,assignCount,createCount;
            
            for (tile = rootTile.childTiles[0]; tile.type != "window"; tile = tile.childTiles[0]);
            
            chrome.windows.create({ type: tile.windowType, url: "about:blank",
                                    left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile), width: 50, height: 50 },
            function(newwin)
            {
                /* this update is necessary to get correct window minimum height */
                
                chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile), width: 50, height: 50 },
                function(win)
                {
                    windowMinWidth = (tile.windowType == "normal") ? win.width-lrbMarginNormal*2 : win.width-lrbMarginPopup*2;
                    windowMinHeight = (tile.windowType == "normal") ? win.height-topMargin-lrbMarginNormal : win.height-topMargin-lrbMarginPopup;
                    
                    chrome.windows.remove(newwin.id);
                });
            });
            
            assignCount = 0;
            createCount = 0;
            
            createTile(rootTile);
            
            function createTile(tile)
            {
                var i,tabId,tabUrl;
                
                if (tile.type == "window")  /* window tile */
                {
                    if (layoutType == 0 || layoutType == 2 || layoutType == 4)
                    {
                        if (assignExist)
                        {
                            if (assignCount < focuswin.tabs.length-selpos)
                            {
                                pos = selpos+assignCount;
                                tile.tabId = focuswin.tabs[pos].id;
                                tile.tabPos = pos;  /* position of assigned tab in parent window */
                                tabId = tile.tabId;
                                tabUrl = null;
                            }
                            else if (assignCount < focuswin.tabs.length)
                            {
                                if (windowCount <= focuswin.tabs.length) pos = selpos-(windowCount-assignCount);
                                else pos = selpos-(focuswin.tabs.length-assignCount);
                                tile.tabId = focuswin.tabs[pos].id;
                                tile.tabPos = pos;  /* position of assigned tab in parent window */
                                tabId = tile.tabId;
                                tabUrl = null;
                            }
                            else
                            {
                                tile.tabId = -1;  /* not yet known */
                                tile.tabPos = -2;  /* generated tab */
                                tabId = null;
                                tabUrl = (defaultPage == "") ? (isFirefox ? null : newtabPage) : defaultPage;
                            }
                            
                            assignCount++;
                        }
                        else
                        {
                            tile.tabId = -1;  /* not yet known */
                            tile.tabPos = -2;  /* generated tab */
                            tabId = null;
                            tabUrl = (defaultPage == "") ? (isFirefox ? null : newtabPage) : defaultPage;
                        }
                    }
                    else  /* layoutType == 1 || layoutType == 3 || layoutType == 5 */
                    {
                        tile.tabId = -1;  /* not yet known */
                        tile.tabPos = -3;  /* tab created by Open Layout */
                        tabId = null;
                        tabUrl = (tile.tabUrl == "") ? (isFirefox ? null : newtabPage) : tile.tabUrl;
                    }
                    
                    chrome.windows.create({ type: tile.windowType, tabId: tabId, url: tabUrl,
                                            left: getTiledWindowLeft(tile)+1, top: getTiledWindowTop(tile)+1,
                                            width: getTiledWindowWidth(tile)-2, height: getTiledWindowHeight(tile)-2 },
                    function(newwin)
                    {
                        chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                          width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                        function(win)  /* Firefox - necessary to correctly position/size both normal (no title) and popup windows */
                        {
                            tile.windowId = newwin.id;
                            tile.tabId = newwin.tabs[0].id;
                            
                            setBadgeForAllTabsInWindow(newwin.id,3);
                            
                            setListenersForAllTabsInWindow(newwin.id,true);
                            
                            setQuickMenuForAllTabsInWindow(newwin.id,showQuickMenu);
                            
                            if (++createCount == windowCount) createFinish();
                        });
                    });
                }
                else  /* split tile */
                {
                    for (i = 0; i < tile.childTiles.length; i++)
                        createTile(tile.childTiles[i]);
                }
            }
        }
        
        function createFinish()
        {
            var tile;
            
            for (tile = rootTile; tile.type != "window"; tile = tile.childTiles[0]);
            
            chrome.windows.update(tile.windowId,{ focused: true },
            function(win)
            {
                updateContextMenuItems(tile.windowId);
                
                startEventHandler();
            });
        }
    });
}

/************************************************************************/

/* Show layout */

function showLayout()
{
    tiledView = true;
    expandedTile = null;
    
    chrome.windows.get(parentWindowId,{ populate: true },
    function(parentwin)
    {
        var count,pos;
        
        setBadgeForAllTabsInWindow(parentWindowId,2);
        
        setListenersForAllTabsInWindow(parentWindowId,false);
        
        setQuickMenuForAllTabsInWindow(parentWindowId,false);
        
        count = 0;
        for (pos = 0; pos < parentwin.tabs.length; pos++)
            if (findTileByTabId(rootTile,parentwin.tabs[pos].id) != null) count++;
        
        if (count == parentwin.tabs.length)  /* prevent parent window closing */
        {
            chrome.tabs.create({ windowId: parentwin.id, url: blankPage },
            function(newtab)
            {
                parentWindowBlankTabId = newtab.id;
                
                showTiles(rootTile);
            });
        }
        else showTiles(rootTile);
        
        function showTiles(tile)
        {
            var showCount;
            
            showCount = 0;
            
            showTile(tile);
            
            function showTile(tile)
            {
                var i,pos,tabId,tabUrl;
                
                if (tile.type == "window")
                {
                    for (pos = 0; pos < parentwin.tabs.length; pos++)
                        if (parentwin.tabs[pos].id == tile.tabId) break;
                    
                    if (pos < parentwin.tabs.length)
                    {
                        if (tile.tabPos >= 0) tile.tabPos = pos;  /* update position of primary tab in parent window */
                        tabId = tile.tabId;
                        tabUrl = null;
                    }
                    else
                    {
                        tile.tabId = -1;  /* not yet known */
                        tile.tabPos = -2;  /* generated tab */
                        tabId = null;
                        tabUrl = (defaultPage == "") ? (isFirefox ? null : newtabPage) : defaultPage;
                    }
                    
                    chrome.windows.create({ type: tile.windowType, tabId: tabId, url: tabUrl,
                                            left: getTiledWindowLeft(tile)+1, top: getTiledWindowTop(tile)+1,
                                            width: getTiledWindowWidth(tile)-2, height: getTiledWindowHeight(tile)-2 },
                    function(newwin)
                    {
                        chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                          width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                        function(win)  /* Firefox - necessary to correctly position/size  normal (no title) and popup windows */
                        {
                            tile.windowId = newwin.id;
                            tile.tabId = newwin.tabs[0].id;
                            
                            setBadgeForAllTabsInWindow(newwin.id,3);
                            
                            setListenersForAllTabsInWindow(newwin.id,true);
                            
                            setQuickMenuForAllTabsInWindow(newwin.id,showQuickMenu);
                            
                            if (++showCount == windowCount) showFinish();
                        });
                    });
                }
                else  /* split tile */
                {
                    for (i = 0; i < tile.childTiles.length; i++)
                        showTile(tile.childTiles[i]);
                }
            }
            
            function showFinish()
            {
                chrome.windows.update(lastFocusedTile.windowId,{ focused: true },
                function(win)
                {
                    updateContextMenuItems(lastFocusedTile.windowId);
                    
                    startEventHandler();
                });
            }
        }
    });
}

/************************************************************************/

/* Hide layout */

function hideLayout(close)
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        tiledView = false;
        expandedTile = null;
        
        closeTabIdList.length = 0;
        autocloseTabIdList.length = 0;
        
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            lastFocusedTile = findTileByWindowId(rootTile,focuswin.id);
            
            if (lastFocusedTile == null)
            {
                for (lastFocusedTile = rootTile; lastFocusedTile.type != "window"; lastFocusedTile = lastFocusedTile.childTiles[0]);
            }
            
            if (parentWindowId == -1)  /* parent window has been closed */
            {
                chrome.windows.create({ type: "normal", url: blankPage,
                                        left: getParentWindowLeft(), top: getParentWindowTop(),
                                        width: getParentWindowWidth(), height: getParentWindowHeight() },
                function(newwin)
                {
                    parentWindowId = newwin.id;
                    parentWindowBlankTabId = newwin.tabs[0].id;
                    
                    hideTiles(rootTile);
                });
            }
            else hideTiles(rootTile);
            
            function hideTiles(tile)
            {
                var hideCount;
                
                hideCount = 0;
                
                hideTile(tile);
                
                function hideTile(tile)
                {
                    chrome.tabs.query({ windowId: tile.windowId },
                    function(tabs)
                    {
                        var i,moveCount,pos,parentPos;
                        
                        tile.windowId = -1;  /* prevents handling in chrome.windows.onRemoved listener */
                        
                        moveCount = 0;
                        
                        if (tile.type == "window")
                        {
                            for (pos = 0; pos < tabs.length; pos++)
                            {
                                if (tabs[pos].id == tile.tabId && tile.tabPos >= 0) parentPos = tile.tabPos;
                                else parentPos = -1;
                                
                                if (tabs[pos].id == tile.tabId && tile.tabPos == -2 && tabs[pos].url == defaultPage)
                                    closeTabIdList.push(tabs[pos].id);
                                
                                if (autoCloseOpen && tabs[pos].id == tile.tabId && tile.tabPos == -3)
                                    autocloseTabIdList.push(tabs[pos].id);
                                
                                if (autoCloseUser && tabs[pos].id != tile.tabId)
                                    autocloseTabIdList.push(tabs[pos].id);
                                
                                chrome.tabs.move(tabs[pos].id,{ windowId: parentWindowId, index: parentPos },
                                function(tab)
                                {
                                    /* Moving last tab out of window, closes window and fires chrome.windows.onRemoved event */
                                    
                                    if (++moveCount == tabs.length) 
                                    {
                                        if (++hideCount == windowCount) hideFinish();
                                    }
                                });
                            }
                        }
                        else  /* split tile */
                        {
                            for (i = 0; i < tile.childTiles.length; i++)
                                hideTile(tile.childTiles[i]);
                        }
                    });
                }
            }
            
            function hideFinish()
            {
                if (parentWindowBlankTabId != -1)
                {
                    chrome.tabs.remove(parentWindowBlankTabId);
                    
                    parentWindowBlankTabId = -1;
                }
                
                window.setTimeout(
                function()
                {
                    while (closeTabIdList.length > 0) chrome.tabs.remove(closeTabIdList.pop());
                },10);  /* avoid console errors for closed tabs */

                chrome.tabs.update(lastFocusedTile.tabId,{ active: true },
                function(tab)
                {
                    chrome.windows.update(parentWindowId,{ focused: true },
                    function(win)
                    {
                        setBadgeForAllTabsInWindow(parentWindowId,1);
                        
                        setListenersForAllTabsInWindow(parentWindowId,false);
                        
                        setQuickMenuForAllTabsInWindow(parentWindowId,false);
                        
                        updateContextMenuItems(parentWindowId);
                        
                        if (close) window.setTimeout(function() { closeLayout(); },100);
                    });
                });
            }
        });
    },deferTimeout);
}

/************************************************************************/

/* Close layout */

function closeLayout()
{
    var windowId;
    
    tiledView = false;
    syncScroll = false;
    expandedTile = null;
    lastFocusedTile = null;
    
    if (parentWindowId != -1) 
    {
        window.setTimeout(
        function()
        {
            while (autocloseTabIdList.length > 0) chrome.tabs.remove(autocloseTabIdList.pop());
        },10);  /* avoid console errors for closed tabs */
        
        setBadgeForAllTabsInWindow(parentWindowId,0);
        
        setListenersForAllTabsInWindow(parentWindowId,false);
        
        setQuickMenuForAllTabsInWindow(parentWindowId,false);
        
        rootTile = null;
        
        windowId = parentWindowId;
        
        parentWindowId = -1;
        
        updateContextMenuItems(windowId);
        
        chrome.storage.local.get("reinstate-state",
        function(object)
        {
            if (object["reinstate-state"] == "maximized")
            {
                chrome.windows.update(windowId,{ state: "maximized" },
                function(win)
                {
                });
            }
        });
    }
    else rootTile = null;
}

/************************************************************************/

/* Refresh Layout */

function refreshLayout()
{
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        lastFocusedTile = findTileByWindowId(rootTile,focuswin.id);
        
        if (lastFocusedTile == null)
        {
            for (lastFocusedTile = rootTile; lastFocusedTile.type != "window"; lastFocusedTile = lastFocusedTile.childTiles[0]);
        }
        
        _refreshLayout();
    });
}
        
function _refreshLayout()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        var updateCount;
        
        expandedTile = null;
        
        updateCount = 0;
        
        updateTile(rootTile);
        
        function updateTile(tile)
        {
            var i;
            
            if (tile.type == "window")
            {
                chrome.windows.update(tile.windowId,{ focused: true,
                                                      left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                      width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                function(win)
                {
                    setBadgeForAllTabsInWindow(win.id,3);
                    
                    setListenersForAllTabsInWindow(win.id,true);
                    
                    setQuickMenuForAllTabsInWindow(win.id,showQuickMenu);
                    
                    if (++updateCount == windowCount)
                    {
                        chrome.windows.update(lastFocusedTile.windowId,{ focused: true },
                        function(win)
                        {
                            updateContextMenuItems(lastFocusedTile.windowId);
                            
                            startEventHandler();
                        });
                    }
                });
            }
            else  /* split tile */
            {
                for (i = 0; i < tile.childTiles.length; i++)
                    updateTile(tile.childTiles[i]);
            }
        }
    },deferTimeout);
}

function _redrawLayout()
{        
    var updateCount;
    
    updateCount = 0;
    
    updateTile(rootTile);
    
    function updateTile(tile)
    {
        var i;
        
        if (tile.type == "window")
        {
            chrome.windows.update(tile.windowId,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                  width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
            function(win)
            {
            });
        }
        else  /* split tile */
        {
            for (i = 0; i < tile.childTiles.length; i++)
                updateTile(tile.childTiles[i]);
        }
    }
}

/************************************************************************/

/* Toggle toolbars */

function toggleToolbars()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        chrome.windows.getLastFocused({ populate: true },
        function(focuswin)
        {
            var tile;
            
            tile = findTileByWindowId(rootTile,focuswin.id);
            
            if (parentWindowId == -1)  /* parent window has been closed */
            {
                chrome.windows.create({ type: "normal", url: blankPage,
                                        left: getParentWindowLeft(), top: getParentWindowTop(),
                                        width: getParentWindowWidth(), height: getParentWindowHeight() },
                function(newwin)
                {
                    parentWindowId = newwin.id;
                    parentWindowBlankTabId = newwin.tabs[0].id;
                    
                    _toggleToolbars(tile);
                });
            }
            else _toggleToolbars(tile);
            
            function _toggleToolbars(tile)
            {
                var moveCount,pos,parentPos;
                
                tile.windowId = -1;  /* prevents handling in chrome.windows.onRemoved listener */
                
                moveCount = 0;
                
                for (pos = 0; pos < focuswin.tabs.length; pos++)
                {
                    if (focuswin.tabs[pos].id == tile.tabId && tile.tabPos >= 0) parentPos = tile.tabPos;
                    else parentPos = -1;
                    
                    chrome.tabs.move(focuswin.tabs[pos].id,{ windowId: parentWindowId, index: parentPos },
                    function(tab)
                    {
                        /* Moving last tab out of window, closes window and fires chrome.windows.onRemoved event */
                        
                        if (++moveCount == focuswin.tabs.length) window.setTimeout(function() { toggleFinish(tile); },100);
                    });
                }
            }
            
            function toggleFinish(tile)
            {
                if (tile.windowType == "normal") tile.windowType = "popup"; else tile.windowType = "normal";
                
                chrome.windows.create({ type: tile.windowType, tabId: tile.tabId,  /* focused: false, */  /* Firefox - does not support focused property */
                                        left: getTiledWindowLeft(tile)+1, top: getTiledWindowTop(tile)+1,
                                        width: getTiledWindowWidth(tile)-2, height: getTiledWindowHeight(tile)-2 },
                function(newwin)
                {
                    chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                      width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                    function(win)  /* Firefox - necessary to correctly position/size both normal (no title) and popup windows */
                    {
                        tile.windowId = newwin.id;
                        tile.tabId = newwin.tabs[0].id;
                        
                        setBadgeForAllTabsInWindow(newwin.id,3);
                        
                        setListenersForAllTabsInWindow(newwin.id,true);
                        
                        setQuickMenuForAllTabsInWindow(newwin.id,showQuickMenu);
                        
                        chrome.windows.update(newwin.id,{ focused: true },
                        function(win)
                        {
                            updateContextMenuItems(newwin.id);
                            
                            if (expandedTile != null) contractTile();
                            else startEventHandler();
                        });
                    });
                });
            }
        });
    },deferTimeout);
}

/************************************************************************/

/* Toggle all toolbars */

function toggleAllToolbars()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        expandedTile = null;
        
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            var windowType;
            
            lastFocusedTile = findTileByWindowId(rootTile,focuswin.id);
            
            if (lastFocusedTile == null)
            {
                for (lastFocusedTile = rootTile; lastFocusedTile.type != "window"; lastFocusedTile = lastFocusedTile.childTiles[0]);
            }
            
            if (lastFocusedTile.windowType == "normal") windowType = "popup"; else windowType = "normal";
            
            if (parentWindowId == -1)  /* parent window has been closed */
            {
                chrome.windows.create({ type: "normal", url: blankPage,
                                        left: getParentWindowLeft(), top: getParentWindowTop(),
                                        width: getParentWindowWidth(), height: getParentWindowHeight() },
                function(newwin)
                {
                    parentWindowId = newwin.id;
                    parentWindowBlankTabId = newwin.tabs[0].id;
                    
                    _toggleAllToolbars(rootTile);
                });
            }
            else _toggleAllToolbars(rootTile);
            
            function _toggleAllToolbars(tile)
            {
                var toggleCount;
                
                toggleCount = 0;
                
                _toggleToolbars(tile);
                
                function _toggleToolbars(tile)
                {
                    chrome.tabs.query({ windowId: tile.windowId },
                    function(tabs)
                    {
                        var i,moveCount,pos,parentPos;
                        
                        tile.windowId = -1;  /* prevents handling in chrome.windows.onRemoved listener */
                        
                        moveCount = 0;
                        
                        if (tile.type == "window")
                        {
                            for (pos = 0; pos < tabs.length; pos++)
                            {
                                if (tabs[pos].id == tile.tabId && tile.tabPos >= 0) parentPos = tile.tabPos;
                                else parentPos = -1;
                                
                                chrome.tabs.move(tabs[pos].id,{ windowId: parentWindowId, index: parentPos },
                                function(tab)
                                {
                                    /* Moving last tab out of window, closes window and fires chrome.windows.onRemoved event */
                                    
                                    if (++moveCount == tabs.length) 
                                    {
                                        tile.windowType = windowType;
                                        
                                        chrome.windows.create({ type: tile.windowType, tabId: tile.tabId,
                                                                left: getTiledWindowLeft(tile)+1, top: getTiledWindowTop(tile)+1,
                                                                width: getTiledWindowWidth(tile)-2, height: getTiledWindowHeight(tile)-2 },
                                        function(newwin)
                                        {
                                            chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                                              width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                                            function(win)  /* Firefox - necessary to correctly position/size both normal (no title) and popup windows */
                                            {
                                                tile.windowId = newwin.id;
                                                tile.tabId = newwin.tabs[0].id;
                                                
                                                setBadgeForAllTabsInWindow(newwin.id,3);
                                                
                                                setListenersForAllTabsInWindow(newwin.id,true);
                                                
                                                setQuickMenuForAllTabsInWindow(newwin.id,showQuickMenu);
                                                
                                                if (++toggleCount == windowCount) toggleAllFinish();
                                            });
                                        });
                                    }
                                });
                            }
                        }
                        else  /* split tile */
                        {
                            for (i = 0; i < tile.childTiles.length; i++)
                                _toggleToolbars(tile.childTiles[i]);
                        }
                    });
                }
            }
            
            function toggleAllFinish()
            {
                chrome.windows.update(lastFocusedTile.windowId,{ focused: true },
                function(win)
                {
                    updateContextMenuItems(win.id);
                    
                    startEventHandler();
                });
            }
        });
    },deferTimeout);
}

/************************************************************************/

/* Add Tile */

function addTile(direction)
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        expandedTile = null;
        
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            var tile,tabUrl;
            var tiles = new Object();
            
            tile = findTileByWindowId(rootTile,focuswin.id);
            
            tiles = addTileOther(tile,direction);
            
            chrome.windows.update(focuswin.id,{ left: getTiledWindowLeft(tiles.modTile), top: getTiledWindowLeft(tiles.modTile),
                                                width: getTiledWindowWidth(tiles.modTile), height: getTiledWindowHeight(tiles.modTile) },
            function(win)
            {
                if (isFirefox) tabUrl = (defaultPage == "") ? null : defaultPage;
                else tabUrl = (defaultPage == "") ? newtabPage : defaultPage;
                
                chrome.windows.create({ type: tiles.newTile.windowType, url: tabUrl,
                                        left: getTiledWindowLeft(tiles.newTile)+1, top: getTiledWindowLeft(tiles.newTile)+1,
                                        width: getTiledWindowWidth(tiles.newTile)-2, height: getTiledWindowHeight(tiles.newTile)-2 },
                function(newwin)
                {
                    chrome.windows.update(newwin.id,{ left: getTiledWindowLeft(tiles.newTile), top: getTiledWindowLeft(tiles.newTile),
                                                      width: getTiledWindowWidth(tiles.newTile), height: getTiledWindowHeight(tiles.newTile) },
                    function(win)  /* Firefox - necessary to correctly position/size both normal (no title) and popup windows */
                    {
                        tiles.newTile.windowId = newwin.id;
                        tiles.newTile.tabPos = -2;  /* generated tab */
                        tiles.newTile.tabId = newwin.tabs[0].id;
                        
                        setBadgeForAllTabsInWindow(newwin.id,3);
                        
                        setListenersForAllTabsInWindow(newwin.id,true);
                        
                        setQuickMenuForAllTabsInWindow(newwin.id,showQuickMenu);
                        
                        lastFocusedTile = tiles.newTile;
                        
                        adjustLayout();
                        
                        _refreshLayout();  /* will restart eventHandler */
                    });
                });
            });
        });
    },deferTimeout);
}

/************************************************************************/

/* Remove Tile */

function removeTile()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        expandedTile = null;
        
        closeTabIdList.length = 0;
        autocloseTabIdList.length = 0;
        
        chrome.windows.getLastFocused({ populate: true },
        function(focuswin)
        {
            var tile;
            
            tile = findTileByWindowId(rootTile,focuswin.id);
            
            if (parentWindowId == -1)  /* parent window has been closed */
            {
                chrome.windows.create({ type: "normal", url: blankPage,
                                        left: getParentWindowLeft(), top: getParentWindowTop(),
                                        width: getParentWindowWidth(), height: getParentWindowHeight() },
                function(newwin)
                {
                    parentWindowId = newwin.id;
                    parentWindowBlankTabId = newwin.tabs[0].id;
                    
                    _removeTile(tile);
                });
            }
            else _removeTile(tile);
            
            function _removeTile(tile)
            {
                var moveCount,pos,parentPos;
                
                tile.windowId = -1;  /* prevents handling in chrome.windows.onRemoved listener */
                
                moveCount = 0;
                
                for (pos = 0; pos < focuswin.tabs.length; pos++)
                {
                    if (focuswin.tabs[pos].id == tile.tabId && tile.tabPos >= 0) parentPos = tile.tabPos;
                    else parentPos = -1;
                    
                    if (focuswin.tabs[pos].id == tile.tabId && tile.tabPos == -2 && focuswin.tabs[pos].url == defaultPage)
                        closeTabIdList.push(focuswin.tabs[pos].id);
                    
                    if (autoCloseOpen && focuswin.tabs[pos].id == tile.tabId && tile.tabPos == -3)
                        autocloseTabIdList.push(focuswin.tabs[pos].id);
                    
                    if (autoCloseUser && focuswin.tabs[pos].id != tile.tabId)
                        autocloseTabIdList.push(focuswin.tabs[pos].id);
                    
                    chrome.tabs.move(focuswin.tabs[pos].id,{ windowId: parentWindowId, index: parentPos },
                    function(tab)
                    {
                        /* Moving last tab out of window, closes window and fires chrome.windows.onRemoved event */
                        
                        if (++moveCount == focuswin.tabs.length)
                        {
                            window.setTimeout(
                            function()
                            {
                                while (closeTabIdList.length > 0) chrome.tabs.remove(closeTabIdList.pop());
                                while (autocloseTabIdList.length > 0) chrome.tabs.remove(autocloseTabIdList.pop());
                            },10);  /* avoid console errors for closed tabs */
                            
                            setBadgeForAllTabsInWindow(parentWindowId,2);
                            
                            setListenersForAllTabsInWindow(parentWindowId,false);
                            
                            setQuickMenuForAllTabsInWindow(parentWindowId,false);
                            
                            lastFocusedTile = removeTileOther(tile);
                            
                            adjustLayout();
                            
                            _refreshLayout();  /* will restart eventHandler */
                        }
                    });
                }
            }
        });
    },deferTimeout);
}

/************************************************************************/

/* Expand Tile */

function expandTile()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        chrome.windows.getLastFocused({ },
        function(focuswin)
        {
            var tile;
            
            tile = expandedTile;
            
            if (expandedTile != null)
            {
                chrome.windows.update(tile.windowId,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                                      width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
                function(win)
                {
                    setQuickMenuForAllTabsInWindow(win.id,showQuickMenu);
                    
                    updateContextMenuItems(win.id);
                });
            }
            
            expandedTile = tile = findTileByWindowId(rootTile,focuswin.id);
            
            chrome.windows.update(focuswin.id,{ left: getParentWindowLeft(), top: getParentWindowTop(),
                                                width: getParentWindowWidth(), height: getParentWindowHeight() },
            function(win)
            {
                setQuickMenuForAllTabsInWindow(win.id,showQuickMenu);
                
                updateContextMenuItems(win.id);
                
                /* Do not start event handler */
            });
        });
    },deferTimeout);
}

/************************************************************************/

/* Contract Tile */

function contractTile()
{
    stopEventHandler();
    
    setTimeout(  /* allow time for tileEventHandler() to complete */
    function()
    {
        var tile;
        
        tile = expandedTile;
        
        expandedTile = null;
        
        chrome.windows.update(tile.windowId,{ left: getTiledWindowLeft(tile), top: getTiledWindowTop(tile),
                                              width: getTiledWindowWidth(tile), height: getTiledWindowHeight(tile) },
        function(win)
        {
            setQuickMenuForAllTabsInWindow(win.id,showQuickMenu);
            
            updateContextMenuItems(win.id);
            
            startEventHandler();
        });
    },deferTimeout);
}

/************************************************************************/

/* Inspect Tile */

function inspectTile(node)
{
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var i,tile;
        
        tile = findTileByWindowId(rootTile,focuswin.id);
        
        if (node == "this") tile = tile;
        else if (node == "before" && tile.parentTile != null && (i = tile.parentTile.childTiles.indexOf(tile)) > 0) tile = tile.parentTile.childTiles[i-1];
        else if (node == "after" && tile.parentTile != null && (i = tile.parentTile.childTiles.indexOf(tile)) < tile.parentTile.childTiles.length-1) tile = tile.parentTile.childTiles[i+1];
        else if (node == "parent" && tile.parentTile != null) tile = tile.parentTile;
        else if (node == "gparent" && tile.parentTile != null && tile.parentTile.parentTile != null) tile = tile.parentTile.parentTile;
        else if (node == "ggparent" && tile.parentTile != null && tile.parentTile.parentTile != null && tile.parentTile.parentTile.parentTile != null) tile = tile.parentTile.parentTile.parentTile;
        else { console.log("Tile does not exist"); return; }
        
        console.log(tile.type + " > ID: " + tile.windowId + " L: " + tile.left + " T: " + tile.top + " W: " + tile.width + " H: " + tile.height + " R: " + tile.ratio.toFixed(4));
    });
}

/************************************************************************/

/* Open Layout */

function openLayout()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        width = 322 + lrbMarginNormal*2;
        height = 338 + topMargin + lrbMarginNormal + ((platformOS != "win") ? -28 : 0);
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.extension.getURL("openlayout.html"),  /* focused: false, */  /* Firefox - does not support focused property */
                                left: left, top: top, width: width, height: height },
        function(newwin)
        {
            chrome.windows.update(newwin.id,{ left: left, top: top },
            function(win)  /* Firefox - necessary to correctly position popup window */
            {
                dialogWindowId = newwin.id;
                
                chrome.tabs.setZoom(newwin.tabs[0].id,1.0);
            });
        });
    });
}

function openLayout_Reply(layoutString)
{
    var j,layoutType;
        
    j = layoutString.indexOf(",");
    layoutType = +layoutString.substr(0,j);
    layoutString = layoutString.substr(j+1);
    
    j = layoutString.indexOf(";");
    windowCount = +layoutString.substr(0,j);
    layoutString = layoutString.substr(j+1);
    
    rootTile = new Tile();
    
    openTile(rootTile);
    
    function openTile(tile)
    {
        var i,j,childCount;
        
        j = layoutString.indexOf(",");
        tile.type = layoutString.substr(0,j);
        layoutString = layoutString.substr(j+1);
        
        if (tile.type == "window")  /* window tile */
        {
            if (layoutType == 0)
            {
                j = layoutString.indexOf(";");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                tile.tabUrl = null;
            }
            else if (layoutType == 1)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over { */
                
                j = layoutString.indexOf("}");
                tile.tabUrl = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over } */
                
                if (isFirefox && (tile.tabUrl.substr(0,6) == "about:" || tile.tabUrl.substr(0,7) == "chrome:" ||
                    tile.tabUrl.substr(0,5) == "data:" || tile.tabUrl.substr(0,11) == "javascript:")) tile.tabUrl = defaultPage;
            }
            else if (layoutType == 2 || layoutType == 4)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                j = layoutString.indexOf(";");
                tile.windowType = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                tile.tabUrl = null;
            }
            else if (layoutType == 3)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                j = layoutString.indexOf(",");
                tile.windowType = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over { */
                
                j = layoutString.indexOf("}");
                tile.tabUrl = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over } */
                
                if (isFirefox && (tile.tabUrl.substr(0,6) == "about:" || tile.tabUrl.substr(0,7) == "chrome:" ||
                    tile.tabUrl.substr(0,5) == "data:" || tile.tabUrl.substr(0,11) == "javascript:")) tile.tabUrl = defaultPage;
            }
            else if (layoutType == 5)
            {
                j = layoutString.indexOf(",");
                tile.ratio = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                j = layoutString.indexOf(",");
                tile.windowType = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over { */
                
                j = layoutString.indexOf("}");
                tile.tabUrl = layoutString.substr(0,j);
                layoutString = layoutString.substr(j+2);  /* step over } */
                
                if (isFirefox && (tile.tabUrl.substr(0,6) == "about:" || tile.tabUrl.substr(0,7) == "chrome:" ||
                    tile.tabUrl.substr(0,5) == "data:" || tile.tabUrl.substr(0,11) == "javascript:")) tile.tabUrl = defaultPage;
                
                j = layoutString.indexOf(",");
                tile.scrollX = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                j = layoutString.indexOf(";");
                tile.scrollY = +layoutString.substr(0,j);
                layoutString = layoutString.substr(j+1);
                
                tile.scrollInit = true;
            }
        }
        else  /* split tile */
        {
            j = layoutString.indexOf(",");
            tile.ratio = +layoutString.substr(0,j);
            layoutString = layoutString.substr(j+1);
            
            j = layoutString.indexOf(";");
            childCount = +layoutString.substr(0,j);
            layoutString = layoutString.substr(j+1);
            
            for (i = 0; i < childCount; i++)
            {
                tile.childTiles[i] = new Tile();
                
                tile.childTiles[i].parentTile = tile;
                
                openTile(tile.childTiles[i]);
            }
        }
    }
    
    adjustLayout();
    
    newLayout(layoutType);
}

/************************************************************************/

/* Save Layout */

function saveLayout()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        width = 322 + lrbMarginNormal*2;
        height = 388 + topMargin + lrbMarginNormal + ((platformOS != "win") ? -26 : 0);
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.extension.getURL("savelayout.html"),  /* focused: false, */  /* Firefox - does not support focused property */
                                left: left, top: top, width: width, height: height },
        function(newwin)
        {
            chrome.windows.update(newwin.id,{ left: left, top: top },
            function(win)  /* Firefox - necessary to correctly position popup window */
            {
                dialogWindowId = newwin.id;
                
                chrome.tabs.setZoom(newwin.tabs[0].id,1.0);
            });
        });
    });
}

function saveLayout_Reply(layoutName,saveTabs)
{
    var layoutType,layoutString;
    var object = new Object();
    
    chrome.tabs.query({ },
    function(tabs)
    {
        layoutType = (saveTabs) ? 5 : 4;
        
        layoutString = layoutType + "," + windowCount + ";";
        
        saveTile(rootTile);
        
        object["layout-string-new"+layoutName] = layoutString;
        
        chrome.storage.local.set(object);
        
        function saveTile(tile)
        {
            var i,tabUrl;
            
            if (tile.type == "window")
            {
                tabUrl = "";
                
                for (i = 0; i < tabs.length; i++)
                    if (tabs[i].id == tile.tabId) tabUrl = tabs[i].url;
                
                layoutString += "window" + "," + tile.ratio.toFixed(4) + "," + tile.windowType;
                if (layoutType == 5)
                {
                    layoutString += "," + "{" + tabUrl + "}";
                    layoutString += "," + tile.scrollX;
                    layoutString += "," + tile.scrollY;
                }
                layoutString += ";";
            }
            else  /* split tile */
            {
                layoutString += tile.type + "," + tile.ratio.toFixed(4) + "," + tile.childTiles.length + ";";
                
                for (i = 0; i < tile.childTiles.length; i++)
                    saveTile(tile.childTiles[i]);
            }
        }
    });
}

/************************************************************************/

/* Delete Layout */

function deleteLayout()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        width = 322 + lrbMarginNormal*2;
        height = 338 + topMargin + lrbMarginNormal + ((platformOS != "win") ? -28 : 0);
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.extension.getURL("deletelayout.html"), /* focused: false, */  /* Firefox - does not support focused property */
                                left: left, top: top, width: width, height: height },
        function(newwin)
        {
            chrome.windows.update(newwin.id,{ left: left, top: top },
            function(win)  /* Firefox - necessary to correctly position popup window */
            {
                dialogWindowId = newwin.id;
                
                chrome.tabs.setZoom(newwin.tabs[0].id,1.0);
            });
        });
    });
}

function deleteLayout_Reply(layoutName)
{
    chrome.storage.local.remove("layout-string-" + layoutName);
}

/************************************************************************/

/* Default Layout */

function defaultLayout()
{
    if (dialogWindowId != -1)
    {
        chrome.windows.remove(dialogWindowId);
        
        dialogWindowId = -1;
    }
    
    chrome.windows.getLastFocused({ },
    function(focuswin)
    {
        var left,top,width,height;
        
        dialogParentWindowId = focuswin.id;
        
        width = 322 + lrbMarginNormal*2;
        height = 338 + topMargin + lrbMarginNormal + ((platformOS != "win") ? -28 : 0);
        
        left = Math.floor(focuswin.left+(focuswin.width-width)/2);
        top = Math.floor(focuswin.top+(focuswin.height-height)/2);
        
        chrome.windows.create({ type: "popup", url: chrome.extension.getURL("defaultlayout.html"), /* focused: false, */  /* Firefox - does not support focused property */
                                left: left, top: top, width: width, height: height },
        function(newwin)
        {
            chrome.windows.update(newwin.id,{ left: left, top: top },
            function(win)  /* Firefox - necessary to correctly position popup window */
            {
                dialogWindowId = newwin.id;
                
                chrome.tabs.setZoom(newwin.tabs[0].id,1.0);
            });
        });
    });
}

function defaultLayout_Reply(layoutName)
{
    chrome.storage.local.set({ "layout-default": layoutName });
}

/************************************************************************/

/* Window resize event handler functions */

function startEventHandler()
{
    if (enableEventHandler && intervalTimer == null)
    {
        intervalTimer = setInterval(function() { tileEventHandler(); },pollInterval);
    }
}

function stopEventHandler()
{
    if (intervalTimer != null)
    {
        clearInterval(intervalTimer);
        intervalTimer = null;
    }
}

function tileEventHandler()
{
    chrome.windows.getLastFocused({ },
    function(win)
    {
        var i,j,tile,rtile,diff,stile,minwidth,minheight;
        
        if (rootTile == null) return;  /* no layout */
        
        tile = findTileByWindowId(rootTile,win.id);
        
        if (tile == null) return;  /* not tiled window */
        
        if (win.state != "normal") return;  /* minimized, maximized or fullscreen window */
        
        if (win.left != getTiledWindowLeft(tile) || win.top != getTiledWindowTop(tile) || 
            win.width != getTiledWindowWidth(tile) || win.height != getTiledWindowHeight(tile))
        {
            if (win.width != getTiledWindowWidth(tile))  /* horizontal resize */
            {
                resizeHoriz = true;
                
                if (win.left != getTiledWindowLeft(tile))  /* resize left edge */
                {
                    resizeBefore = true;
                    
                    diff = win.width-getTiledWindowWidth(tile);
                    
                    tile.left -= diff;
                    tile.width += diff;
                }
                else  /* resize right edge */
                {
                    resizeBefore = false;
                    
                    diff = win.width-getTiledWindowWidth(tile);
                    
                    tile.width += diff;
                }
            }
            else if (win.height != getTiledWindowHeight(tile))  /* vertical resize */
            {
                resizeHoriz = false;
                
                if (win.top != getTiledWindowTop(tile))  /* resize top edge */
                {
                    resizeBefore = true;
                    
                    diff = win.height-getTiledWindowHeight(tile);
                    
                    tile.top -= diff;
                    tile.height += diff;
                }
                else  /* resize bottom edge */
                {
                    resizeBefore = false;
                    
                    diff = win.height-getTiledWindowHeight(tile);
                    
                    tile.height += diff;
                }
            }
            
            idleCounter = 0;
        }
        else if (idleCounter != -1 && ++idleCounter >= 3)  /* 300 ms */
        {
            idleCounter = -1;
            
            rtile = tile;
            
            if (resizeHoriz)
            {
                if (tile.parentTile.type == "hsplit") diff = tile.width-tile.parentTile.width;
                else diff = tile.width-Math.round(tile.parentTile.width*tile.ratio);
                
                if (resizeBefore) tile.left -= diff;
                tile.width -= diff;
                
                while (rtile.parentTile != null)  /* not at root */
                {
                    if (rtile.parentTile.type != "vsplit") rtile = rtile.parentTile;
                    else if (rtile.parentTile.childTiles.indexOf(rtile) == (resizeBefore ? 0 : rtile.parentTile.childTiles.length-1)) rtile = rtile.parentTile;
                    else break;
                }
                
                if (rtile.parentTile != null)  /* not at root */
                {
                    i = rtile.parentTile.childTiles.indexOf(rtile);
                    
                    if (resizeBefore)  /* resize left edge */
                    {
                        stile = rtile.parentTile.childTiles[i-1];
                        
                        if (rtile.width+diff < getTileMinWidth(rtile)) diff = getTileMinWidth(rtile)-rtile.width;
                        else if (stile.width-diff < getTileMinWidth(stile)) diff = stile.width-getTileMinWidth(stile);
                        
                        rtile.left -= diff;
                    }
                    else  /* resize right edge */
                    {
                        stile = rtile.parentTile.childTiles[i+1];
                        
                        if (rtile.width+diff < getTileMinWidth(rtile)) diff = getTileMinWidth(rtile)-rtile.width;
                        else if (stile.width-diff < getTileMinWidth(stile)) diff = stile.width-getTileMinWidth(stile);
                        
                        stile.left += diff;
                    }
                    
                    rtile.width += diff;
                    stile.width -= diff;
                    
                    rtile.ratio = rtile.width/rtile.parentTile.width;
                    stile.ratio = stile.width/stile.parentTile.width;
               }
            }
            else
            {
                if (tile.parentTile.type == "vsplit") diff = tile.height-tile.parentTile.height;
                else diff = tile.height-Math.round(tile.parentTile.height*tile.ratio);
                
                if (resizeBefore) tile.top -= diff;
                tile.height -= diff;
                
                while (rtile.parentTile != null)  /* not at root */
                {
                    if (rtile.parentTile.type != "hsplit") rtile = rtile.parentTile;
                    else if (rtile.parentTile.childTiles.indexOf(rtile) == (resizeBefore ? 0 : rtile.parentTile.childTiles.length-1)) rtile = rtile.parentTile;
                    else break;
                }
                
                if (rtile.parentTile != null)  /* not at root */
                {
                    i = rtile.parentTile.childTiles.indexOf(rtile);
                    
                    if (resizeBefore)  /* resize top edge */
                    {
                        stile = rtile.parentTile.childTiles[i-1];
                        
                        if (rtile.height+diff < getTileMinHeight(rtile)) diff = getTileMinHeight(rtile)-rtile.height;
                        else if (stile.height-diff < getTileMinHeight(stile)) diff = stile.height-getTileMinHeight(stile);
                        
                        rtile.top -= diff;
                    }
                    else  /* resize bottom edge */
                    {
                        stile = rtile.parentTile.childTiles[i+1];
                        
                        if (rtile.height+diff < getTileMinHeight(rtile)) diff = getTileMinHeight(rtile)-rtile.height;
                        else if (stile.height-diff < getTileMinHeight(stile)) diff = stile.height-getTileMinHeight(stile);
                        
                        stile.top += diff;
                    }
                    
                    rtile.height += diff;
                    stile.height -= diff;
                    
                    rtile.ratio = rtile.height/rtile.parentTile.height;
                    stile.ratio = stile.height/stile.parentTile.height;
                }
            }
            
            adjustLayout();
            
            _redrawLayout();
        }
    });
}

/************************************************************************/

/* Check for sendMessage errors */

function checkError()
{
    if (chrome.runtime.lastError == null) ;
    else if (chrome.runtime.lastError.message == "Could not establish connection. Receiving end does not exist.") ;  /* Chrome & Firefox - ignore */
    else if (chrome.runtime.lastError.message == "The message port closed before a response was received.") ;  /* Chrome - ignore */
    else if (chrome.runtime.lastError.message == "Message manager disconnected") ;  /* Firefox - ignore */
    else console.log("Tile Tabs WE - " + chrome.runtime.lastError.message);
}

/************************************************************************/

/* Display alert notification */

function alertNotify(message)
{
    chrome.notifications.create("alert",{ type: "basic", iconUrl: "icon32.png", title: "TILE TABS WE", message: "" + message });
}

/************************************************************************/

/* Display debug notification */

function debugNotify(message)
{
    chrome.notifications.create("debug",{ type: "basic", iconUrl: "icon32.png", title: "TILE TABS WE - DEBUG", message: "" + message });
}

/************************************************************************/
